try:
    # Import C implementation
    from .c import *
except ImportError:
    try:
        # Import CFFI implementation
        from .cffi import *
    except ImportError:
        raise ImportError(
            "\n\npyzstd module: Can't import compiled .so/.pyd file.\n"
            "1, If pyzstd module is dynamically linked to zstd library: Make sure\n"
            "   not to remove zstd library, and the run-time zstd library's version\n"
            "   can't be lower than that at compile-time; On Windows, the directory\n"
            "   that has libzstd.dll should be added by os.add_dll_directory() function.\n"
            "2, Please install pyzstd module through pip, to ensure that compiled\n"
            "   .so/.pyd file matches the architecture/OS/Python.\n")
from .zstdfile import *
from .seekable_zstdfile import *

__version__ = '0.15.9'

__doc__ = '''\
Python bindings to Zstandard (zstd) compression library, the API style is
similar to Python's bz2/lzma/zlib modules.

Command line interface of this module: python -m pyzstd --help

Documentation: https://pyzstd.readthedocs.io
GitHub: https://github.com/animalize/pyzstd
PyPI: https://pypi.org/project/pyzstd'''

__all__ = ('ZstdCompressor', 'RichMemZstdCompressor',
           'ZstdDecompressor', 'EndlessZstdDecompressor',
           'CParameter', 'DParameter', 'Strategy', 'ZstdError',
           'compress', 'richmem_compress', 'decompress',
           'compress_stream', 'decompress_stream',
           'ZstdDict', 'train_dict', 'finalize_dict',
           'get_frame_info', 'get_frame_size', 'ZstdFile', 'open',
           'zstd_version', 'zstd_version_info',
           'zstd_support_multithread', 'compressionLevel_values',
           'SeekableZstdFile', 'SeekableFormatError')


zstd_support_multithread = (CParameter.nbWorkers.bounds() != (0, 0))


def compress(data, level_or_option=None, zstd_dict=None):
    """Compress a block of data, return a bytes object.

    Compressing b'' will get an empty content frame (9 bytes or more).

    Parameters
    data:            A bytes-like object, data to be compressed.
    level_or_option: When it's an int object, it represents compression level.
                     When it's a dict object, it contains advanced compression
                     parameters.
    zstd_dict:       A ZstdDict object, pre-trained dictionary for compression.
    """
    comp = ZstdCompressor(level_or_option, zstd_dict)
    return comp.compress(data, ZstdCompressor.FLUSH_FRAME)


def richmem_compress(data, level_or_option=None, zstd_dict=None):
    """Compress a block of data, return a bytes object.

    Use rich memory mode, it's faster than compress() in some cases, but
    allocates more memory.

    Compressing b'' will get an empty content frame (9 bytes or more).

    Parameters
    data:            A bytes-like object, data to be compressed.
    level_or_option: When it's an int object, it represents compression level.
                     When it's a dict object, it contains advanced compression
                     parameters.
    zstd_dict:       A ZstdDict object, pre-trained dictionary for compression.
    """
    comp = RichMemZstdCompressor(level_or_option, zstd_dict)
    return comp.compress(data)


def _nbytes(dat):
    if isinstance(dat, (bytes, bytearray)):
        return len(dat)
    with memoryview(dat) as mv:
        return mv.nbytes


def train_dict(samples, dict_size):
    """Train a zstd dictionary, return a ZstdDict object.

    Parameters
    samples:   An iterable of samples, a sample is a bytes-like object
               represents a file.
    dict_size: The dictionary's maximum size, in bytes.
    """
    # Check argument's type
    if not isinstance(dict_size, int):
        raise TypeError('dict_size argument should be an int object.')

    # Prepare data
    chunks = []
    chunk_sizes = []
    for chunk in samples:
        chunks.append(chunk)
        chunk_sizes.append(_nbytes(chunk))

    chunks = b''.join(chunks)
    if not chunks:
        raise ValueError("The samples are empty content, can't train dictionary.")

    # samples_bytes: samples be stored concatenated in a single flat buffer.
    # samples_size_list: a list of each sample's size.
    # dict_size: size of the dictionary, in bytes.
    dict_content = _train_dict(chunks, chunk_sizes, dict_size)

    return ZstdDict(dict_content)


def finalize_dict(zstd_dict, samples, dict_size, level):
    """Finalize a zstd dictionary, return a ZstdDict object.

    Given a custom content as a basis for dictionary, and a set of samples,
    finalize dictionary by adding headers and statistics according to the zstd
    dictionary format.

    You may compose an effective dictionary content by hand, which is used as
    basis dictionary, and use some samples to finalize a dictionary. The basis
    dictionary can be a "raw content" dictionary, see is_raw parameter in
    ZstdDict.__init__ method.

    Parameters
    zstd_dict: A ZstdDict object, basis dictionary.
    samples:   An iterable of samples, a sample is a bytes-like object
               represents a file.
    dict_size: The dictionary's maximum size, in bytes.
    level:     The compression level expected to use in production. The
               statistics for each compression level differ, so tuning the
               dictionary for the compression level can help quite a bit.
    """
    if zstd_version_info < (1, 4, 5):
        msg = ("This function only available when the underlying zstd "
               "library's version is greater than or equal to v1.4.5, "
               "the current underlying zstd library's version is v%s.") % zstd_version
        raise NotImplementedError(msg)

    # Check arguments' type
    if not isinstance(zstd_dict, ZstdDict):
        raise TypeError('zstd_dict argument should be a ZstdDict object.')
    if not isinstance(dict_size, int):
        raise TypeError('dict_size argument should be an int object.')
    if not isinstance(level, int):
        raise TypeError('level argument should be an int object.')

    # Prepare data
    chunks = []
    chunk_sizes = []
    for chunk in samples:
        chunks.append(chunk)
        chunk_sizes.append(_nbytes(chunk))

    chunks = b''.join(chunks)
    if not chunks:
        raise ValueError("The samples are empty content, can't finalize dictionary.")

    # custom_dict_bytes: existing dictionary.
    # samples_bytes: samples be stored concatenated in a single flat buffer.
    # samples_size_list: a list of each sample's size.
    # dict_size: maximal size of the dictionary, in bytes.
    # compression_level: compression level expected to use in production.
    dict_content = _finalize_dict(zstd_dict.dict_content,
                                  chunks, chunk_sizes,
                                  dict_size, level)

    return ZstdDict(dict_content)



from .common import m, ffi, ZstdError, _new_nonzero, _nbytes, \
                    _set_c_parameters, _set_d_parameters, \
                    _write_to_fp, _ZSTD_DStreamSizes, \
                    _set_zstd_error, _ErrorType
from .dict import _load_c_dict, _load_d_dict
from .output_buffer import _BlocksOutputBuffer

_ZSTD_DStreamOutSize = _ZSTD_DStreamSizes[1]

class ZstdFileReader:
    def __init__(self, fp, zstd_dict, option, read_size):
        if read_size <= 0:
            raise ValueError("read_size argument should > 0")
        self._read_size = read_size

        # File states, the last three are public attributes.
        self._fp = fp
        self.eof = False
        self.pos = 0     # Decompressed position
        self.size = -1   # File size, -1 means unknown.

        # Decompression states
        self._needs_input = True
        self._at_frame_edge = True

        # Input state, need to be initialized with 0.
        self._in_buf = ffi.new("ZSTD_inBuffer *")
        if self._in_buf == ffi.NULL:
            raise MemoryError
        # Output state
        self._out_buf = _new_nonzero("ZSTD_outBuffer *")
        if self._out_buf == ffi.NULL:
            raise MemoryError
        # Lazy create forward output buffer
        self._tmp_output = ffi.NULL

        # Decompression context
        self._dctx = m.ZSTD_createDCtx()
        if self._dctx == ffi.NULL:
            raise ZstdError("Unable to create ZSTD_DCtx instance.")

        # Load dictionary to decompression context
        if zstd_dict is not None:
            _load_d_dict(self._dctx, zstd_dict)
            self.__dict = zstd_dict

        # Set option to decompression context
        if option is not None:
            _set_d_parameters(self._dctx, option)

    def __del__(self):
        try:
            m.ZSTD_freeDCtx(self._dctx)
            self._dctx = ffi.NULL
        except AttributeError:
            pass

    def _decompress_into(self, out_b, fill_full):
        # Return
        if self.eof or out_b.size == out_b.pos:
            return

        in_b = self._in_buf
        orig_pos = out_b.pos
        while True:
            if in_b.size == in_b.pos and self._needs_input:
                # Read
                self._in_dat = self._fp.read(self._read_size)
                # EOF
                if not self._in_dat:
                    if self._at_frame_edge:
                        self.eof = True
                        self.pos += out_b.pos - orig_pos
                        self.size = self.pos
                        return
                    else:
                        raise EOFError("Compressed file ended before the "
                                       "end-of-stream marker was reached")
                in_b.src = ffi.from_buffer(self._in_dat)
                in_b.size = _nbytes(self._in_dat)
                in_b.pos = 0

            # Decompress
            zstd_ret = m.ZSTD_decompressStream(self._dctx, out_b, in_b)
            if m.ZSTD_isError(zstd_ret):
                _set_zstd_error(_ErrorType.ERR_DECOMPRESS, zstd_ret)

            # Set flags
            if zstd_ret == 0:
                self._needs_input = True
                self._at_frame_edge = True
            else:
                self._needs_input = (out_b.size != out_b.pos)
                self._at_frame_edge = False

            if fill_full:
                if out_b.size != out_b.pos:
                    continue
                else:
                    self.pos += out_b.pos - orig_pos
                    return
            else:
                if out_b.pos != orig_pos:
                    self.pos += out_b.pos - orig_pos
                    return

    def readinto(self, b):
        out_b = self._out_buf
        out_b.dst = ffi.from_buffer(b)
        out_b.size = _nbytes(b)
        out_b.pos = 0

        self._decompress_into(out_b, False)
        return out_b.pos

    def readall(self):
        out_b = self._out_buf
        out = _BlocksOutputBuffer()
        if self.size >= 0:
            # Known file size
            out.initWithSize(out_b, -1, self.size - self.pos)
        else:
            # Unknown file size
            out.initAndGrow(out_b, -1)

        while True:
            self._decompress_into(out_b, True)
            if self.eof:
                # Finished
                return out.finish(out_b)
            if out_b.size == out_b.pos:
                # Grow output buffer
                out.grow(out_b)

    # If obj is None, forward to EOF.
    # If obj <= 0, do nothing.
    def forward(self, offset):
        # Lazy create forward output buffer
        if self._tmp_output == ffi.NULL:
            # ZSTD_outBuffer struct
            self._out_tmp = _new_nonzero("ZSTD_outBuffer *")
            if self._out_tmp == ffi.NULL:
                raise MemoryError
            # Forward output buffer
            self._tmp_output = _new_nonzero("char[]", _ZSTD_DStreamOutSize)
            if self._tmp_output == ffi.NULL:
                raise MemoryError
            # ZSTD_outBuffer.dst
            self._out_tmp.dst = self._tmp_output
        out_b = self._out_tmp

        # Forward to EOF
        if offset is None:
            out_b.size = _ZSTD_DStreamOutSize
            while True:
                out_b.pos = 0
                self._decompress_into(out_b, True)
                if self.eof:
                    return

        # Forward to offset
        while offset > 0:
            out_b.size = min(_ZSTD_DStreamOutSize, offset)
            out_b.pos = 0
            self._decompress_into(out_b, True)

            if self.eof:
                return
            offset -= out_b.pos

    def reset_session(self):
        # Reset decompression states
        self._needs_input = True
        self._at_frame_edge = True
        self._in_buf.size = 0
        self._in_buf.pos = 0

        # Resetting session never fail
        m.ZSTD_DCtx_reset(self._dctx, m.ZSTD_reset_session_only)

class ZstdFileWriter:
    def __init__(self, fp, level_or_option, zstd_dict, write_size):
        # File object
        self._fp = fp
        self._fp_has_flush = hasattr(fp, "flush")

        # States
        self._last_mode = m.ZSTD_e_end
        self._use_multithread = False
        level = 0  # 0 means use zstd's default compression level

        # Write buffer
        if write_size <= 0:
            raise ValueError("write_size argument should > 0")
        self._write_buffer_size = write_size

        self._write_buffer = _new_nonzero("char[]", write_size)
        if self._write_buffer == ffi.NULL:
            raise MemoryError

        # Singleton buffer objects
        self._in_buf = _new_nonzero("ZSTD_inBuffer *")
        if self._in_buf == ffi.NULL:
            raise MemoryError

        self._out_buf = _new_nonzero("ZSTD_outBuffer *")
        if self._out_buf == ffi.NULL:
            raise MemoryError

        self._out_mv = memoryview(ffi.buffer(self._write_buffer))

        # Compression context
        self._cctx = m.ZSTD_createCCtx()
        if self._cctx == ffi.NULL:
            raise ZstdError("Unable to create ZSTD_CCtx instance.")

        # Set compressLevel/option to compression context
        if level_or_option is not None:
            level, self._use_multithread = \
                  _set_c_parameters(self._cctx, level_or_option)

        # Load dictionary to compression context
        if zstd_dict is not None:
            _load_c_dict(self._cctx, zstd_dict, level)
            self.__dict = zstd_dict

    def __del__(self):
        try:
            m.ZSTD_freeCCtx(self._cctx)
            self._cctx = ffi.NULL
        except AttributeError:
            pass

    def write(self, data):
        # Output size
        output_size = 0

        # Input buffer
        in_b = self._in_buf
        in_b.src = ffi.from_buffer(data)
        in_b.size = _nbytes(data)
        in_b.pos = 0

        # Output buffer, out.pos will be set later.
        out_b = self._out_buf
        out_b.dst = self._write_buffer
        out_b.size = self._write_buffer_size

        # State
        self._last_mode = m.ZSTD_e_continue

        while True:
            # Output position
            out_b.pos = 0

            # Compress
            if not self._use_multithread:
                zstd_ret = m.ZSTD_compressStream2(self._cctx, out_b, in_b,
                                                  m.ZSTD_e_continue)
            else:
                while True:
                    zstd_ret = m.ZSTD_compressStream2(self._cctx, out_b, in_b,
                                                      m.ZSTD_e_continue)
                    if (out_b.pos == out_b.size
                           or in_b.pos == in_b.size
                           or m.ZSTD_isError(zstd_ret)):
                        break

            if m.ZSTD_isError(zstd_ret):
                _set_zstd_error(_ErrorType.ERR_COMPRESS, zstd_ret)

            # Accumulate output bytes
            output_size += out_b.pos

            # Write output to fp
            _write_to_fp("self._fp.write()", self._fp,
                         self._out_mv, out_b)

            # Finished
            if not self._use_multithread:
                # Single-thread compression + .CONTINUE mode
                if zstd_ret == 0:
                    break
            else:
                # Multi-thread compression + .CONTINUE mode
                if in_b.size == in_b.pos and \
                   out_b.size != out_b.pos:
                    break

        return (in_b.size, output_size)

    def flush(self, mode):
        # Mode argument
        if mode not in (m.ZSTD_e_flush, m.ZSTD_e_end):
            msg = ("mode argument wrong value, it should be "
                   "ZstdFile.FLUSH_BLOCK or ZstdFile.FLUSH_FRAME.")
            raise ValueError(msg)

        # Don't generate empty content frame
        if mode == self._last_mode:
            return (0, 0)

        # Output size
        output_size = 0

        # Input buffer
        in_b = self._in_buf
        in_b.src = self._write_buffer
        in_b.size = 0
        in_b.pos = 0

        # Output buffer, out.pos will be set later.
        out_b = self._out_buf
        out_b.dst = self._write_buffer
        out_b.size = self._write_buffer_size

        # State
        self._last_mode = mode

        while True:
            # Output position
            out_b.pos = 0

            # Compress
            zstd_ret = m.ZSTD_compressStream2(self._cctx, out_b, in_b, mode)
            if m.ZSTD_isError(zstd_ret):
                _set_zstd_error(_ErrorType.ERR_COMPRESS, zstd_ret)

            # Accumulate output bytes
            output_size += out_b.pos

            # Write output to fp
            _write_to_fp("self._fp.write()", self._fp,
                         self._out_mv, out_b)

            # Finished
            if zstd_ret == 0:
                break

        # Flush
        if self._fp_has_flush:
            self._fp.flush()

        return (0, output_size)



from threading import Lock

from .common import m, ffi, ZstdError, \
                    _new_nonzero, _set_d_parameters, \
                    _set_zstd_error, _ErrorType
from .dict import _load_d_dict
from .output_buffer import _BlocksOutputBuffer

_TYPE_DEC         = 0
_TYPE_ENDLESS_DEC = 1

class _Decompressor:
    def __init__(self, zstd_dict=None, option=None):
        self._lock = Lock()
        self._needs_input = True
        self._input_buffer = ffi.NULL
        self._input_buffer_size = 0
        self._in_begin = 0
        self._in_end = 0

        self._singleton_in_buf = _new_nonzero("ZSTD_inBuffer *")
        if self._singleton_in_buf == ffi.NULL:
            raise MemoryError

        self._singleton_out_buf = _new_nonzero("ZSTD_outBuffer *")
        if self._singleton_out_buf == ffi.NULL:
            raise MemoryError

        # Decompression context
        self._dctx = m.ZSTD_createDCtx()
        if self._dctx == ffi.NULL:
            raise ZstdError("Unable to create ZSTD_DCtx instance.")

        # Load dictionary to compression context
        if zstd_dict is not None:
            _load_d_dict(self._dctx, zstd_dict)
            self.__dict = zstd_dict

        # Set compressLevel/option to compression context
        if option is not None:
            _set_d_parameters(self._dctx, option)

    def __del__(self):
        try:
            m.ZSTD_freeDCtx(self._dctx)
            self._dctx = ffi.NULL
        except AttributeError:
            pass

    @property
    def needs_input(self):
        """If the max_length output limit in .decompress() method has been reached, and
        the decompressor has (or may has) unconsumed input data, it will be set to
        False. In this case, pass b'' to .decompress() method may output further data.
        """
        return self._needs_input

    def _decompress_impl(self, in_buf, max_length, initial_size):
        # The first AFE check for setting .at_frame_edge flag, search "AFE" in
        # decompressor.c to see details.
        if self._type == _TYPE_ENDLESS_DEC:
            if self._at_frame_edge and in_buf.pos == in_buf.size:
                return b""

        # Output buffer
        out_buf = self._singleton_out_buf
        out = _BlocksOutputBuffer()
        if initial_size >= 0:
            out.initWithSize(out_buf, max_length, initial_size)
        else:
            out.initAndGrow(out_buf, max_length)

        while True:
            # Decompress
            zstd_ret = m.ZSTD_decompressStream(self._dctx, out_buf, in_buf)
            if m.ZSTD_isError(zstd_ret):
                _set_zstd_error(_ErrorType.ERR_DECOMPRESS, zstd_ret)

            # Set .eof/.af_frame_edge flag
            if self._type == _TYPE_DEC:
                # ZstdDecompressor class stops when a frame is decompressed
                if zstd_ret == 0:
                    self._eof = True
                    break
            else:
                # EndlessZstdDecompressor class supports multiple frames
                self._at_frame_edge = (zstd_ret == 0)

                # The second AFE check for setting .at_frame_edge flag, search
                # "AFE" in decompressor.c to see details.
                if self._at_frame_edge and in_buf.pos == in_buf.size:
                    break

            # Need to check out before in. Maybe zstd's internal buffer still has
            # a few bytes can be output, grow the buffer and continue.
            if out_buf.pos == out_buf.size:
                # Output buffer exhausted

                # Output buffer reached max_length
                if out.reachedMaxLength(out_buf):
                    break

                # Grow output buffer
                out.grow(out_buf)
            elif in_buf.pos == in_buf.size:
                # Finished
                break

        return out.finish(out_buf)

    def _stream_decompress(self, data, max_length=-1):
        self._lock.acquire()
        try:
            initial_buffer_size = -1
            in_buf = self._singleton_in_buf

            if self._type == _TYPE_DEC:
                # Check .eof flag
                if self._eof:
                    raise EOFError("Already at the end of a zstd frame.")
            else:
                # Fast path for the first frame
                if self._at_frame_edge and self._in_begin == self._in_end:
                    # Read decompressed size
                    decompressed_size = m.ZSTD_getFrameContentSize(ffi.from_buffer(data),
                                                                   len(data))

                    # Use ZSTD_findFrameCompressedSize() to check complete frame,
                    # prevent allocating too much memory for small input chunk.
                    if (decompressed_size not in (m.ZSTD_CONTENTSIZE_UNKNOWN,
                                                  m.ZSTD_CONTENTSIZE_ERROR) \
                          and \
                          not m.ZSTD_isError(m.ZSTD_findFrameCompressedSize(ffi.from_buffer(data),
                                                                            len(data))) ):
                        initial_buffer_size = decompressed_size

            # Prepare input buffer w/wo unconsumed data
            if self._in_begin == self._in_end:
                # No unconsumed data
                use_input_buffer = False

                in_buf.src = ffi.from_buffer(data)
                in_buf.size = len(data)
                in_buf.pos = 0
            elif len(data) == 0:
                # Has unconsumed data, fast path for b"".
                use_input_buffer = True

                in_buf.src = self._input_buffer + self._in_begin
                in_buf.size = self._in_end - self._in_begin
                in_buf.pos = 0
            else:
                # Has unconsumed data
                use_input_buffer = True

                # Unconsumed data size in input_buffer
                used_now = self._in_end - self._in_begin
                # Number of bytes we can append to input buffer
                avail_now = self._input_buffer_size - self._in_end
                # Number of bytes we can append if we move existing
                # contents to beginning of buffer
                avail_total = self._input_buffer_size - used_now

                if avail_total < len(data):
                    new_size = used_now + len(data)
                    # Allocate with new size
                    tmp = _new_nonzero("char[]", new_size)
                    if tmp == ffi.NULL:
                        raise MemoryError

                    # Copy unconsumed data to the beginning of new buffer
                    ffi.memmove(tmp,
                                self._input_buffer+self._in_begin,
                                used_now)

                    # Switch to new buffer
                    self._input_buffer = tmp
                    self._input_buffer_size = new_size

                    # Set begin & end position
                    self._in_begin = 0
                    self._in_end = used_now
                elif avail_now < len(data):
                    # Move unconsumed data to the beginning
                    ffi.memmove(self._input_buffer,
                                self._input_buffer+self._in_begin,
                                used_now)

                    # Set begin & end position
                    self._in_begin = 0
                    self._in_end = used_now

                # Copy data to input buffer
                ffi.memmove(self._input_buffer+self._in_end,
                            ffi.from_buffer(data), len(data))
                self._in_end += len(data)

                in_buf.src = self._input_buffer + self._in_begin
                in_buf.size = used_now + len(data)
                in_buf.pos = 0
            # Now in_buf.pos == 0

            ret = self._decompress_impl(in_buf, max_length, initial_buffer_size)

            # Unconsumed input data
            if in_buf.pos == in_buf.size:
                if self._type == _TYPE_DEC:
                    if len(ret) == max_length or self._eof:
                        self._needs_input = False
                    else:
                        self._needs_input = True
                else:
                    if len(ret) == max_length and not self._at_frame_edge:
                        self._needs_input = False
                    else:
                        self._needs_input = True

                if use_input_buffer:
                    # Clear input_buffer
                    self._in_begin = 0
                    self._in_end = 0
            else:
                data_size = in_buf.size - in_buf.pos

                self._needs_input = False
                if self._type == _TYPE_ENDLESS_DEC:
                    self._at_frame_edge = False

                if not use_input_buffer:
                    # Discard buffer if it's too small
                    if (self._input_buffer == ffi.NULL
                          or self._input_buffer_size < data_size):
                        # Create new buffer
                        self._input_buffer = _new_nonzero("char[]", data_size)
                        if self._input_buffer == ffi.NULL:
                            self._input_buffer_size = 0
                            raise MemoryError
                        # Set buffer size
                        self._input_buffer_size = data_size

                    # Copy unconsumed data
                    ffi.memmove(self._input_buffer, in_buf.src+in_buf.pos, data_size)
                    self._in_begin = 0
                    self._in_end = data_size
                else:
                    # Use input buffer
                    self._in_begin += in_buf.pos

            return ret
        except:
            # Reset decompressor's states/session
            self.__reset_session()
            raise
        finally:
            self._lock.release()

    def __reset_session(self):
        # Reset variables
        self._in_begin = 0
        self._in_end = 0

        self._needs_input = True
        if self._type == _TYPE_DEC:
            self._eof = False
            self._unused_data = ffi.NULL
        else:
            self._at_frame_edge = True

        # Resetting session never fail
        m.ZSTD_DCtx_reset(self._dctx, m.ZSTD_reset_session_only)

    def _reset_session(self):
        """This is an undocumented method. Reset decompressor's states/session, don't
        reset parameters and dictionary.
        """
        with self._lock:
            self.__reset_session()

    def __reduce__(self):
        msg = "Cannot pickle %s object." % type(self)
        raise TypeError(msg)

class ZstdDecompressor(_Decompressor):
    """A streaming decompressor, it stops after a frame is decompressed.
    Thread-safe at method level."""

    def __init__(self, zstd_dict=None, option=None):
        """Initialize a ZstdDecompressor object.

        Parameters
        zstd_dict: A ZstdDict object, pre-trained zstd dictionary.
        option:    A dict object that contains advanced decompression parameters.
        """
        super().__init__(zstd_dict, option)
        self._eof = False
        self._unused_data = ffi.NULL
        self._type = _TYPE_DEC

    def decompress(self, data, max_length=-1):
        """Decompress data, return a chunk of decompressed data if possible, or b''
        otherwise.

        It stops after a frame is decompressed.

        Parameters
        data:       A bytes-like object, zstd data to be decompressed.
        max_length: Maximum size of returned data. When it is negative, the size of
                    output buffer is unlimited. When it is nonnegative, returns at
                    most max_length bytes of decompressed data.
        """
        return self._stream_decompress(data, max_length)

    @property
    def eof(self):
        """True means the end of the first frame has been reached. If decompress data
        after that, an EOFError exception will be raised."""
        return self._eof

    @property
    def unused_data(self):
        """A bytes object. When ZstdDecompressor object stops after a frame is
        decompressed, unused input data after the frame. Otherwise this will be b''."""
        with self._lock:
            if not self._eof:
                return b""
            else:
                if self._unused_data == ffi.NULL:
                    if self._input_buffer == ffi.NULL:
                        self._unused_data = b""
                    else:
                        self._unused_data = \
                            ffi.buffer(self._input_buffer)[self._in_begin:self._in_end]
                return self._unused_data

class EndlessZstdDecompressor(_Decompressor):
    """A streaming decompressor, accepts multiple concatenated frames.
    Thread-safe at method level."""

    def __init__(self, zstd_dict=None, option=None):
        """Initialize an EndlessZstdDecompressor object.

        Parameters
        zstd_dict: A ZstdDict object, pre-trained zstd dictionary.
        option:    A dict object that contains advanced decompression parameters.
        """
        super().__init__(zstd_dict, option)
        self._at_frame_edge = True
        self._type = _TYPE_ENDLESS_DEC

    def decompress(self, data, max_length=-1):
        """Decompress data, return a chunk of decompressed data if possible, or b''
        otherwise.

        Parameters
        data:       A bytes-like object, zstd data to be decompressed.
        max_length: Maximum size of returned data. When it is negative, the size of
                    output buffer is unlimited. When it is nonnegative, returns at
                    most max_length bytes of decompressed data.
        """
        return self._stream_decompress(data, max_length)

    @property
    def at_frame_edge(self):
        """True when both the input and output streams are at a frame edge, means a
        frame is completely decoded and fully flushed, or the decompressor just be
        initialized.

        This flag could be used to check data integrity in some cases.
        """
        return self._at_frame_edge

def decompress(data, zstd_dict=None, option=None):
    """Decompress a zstd data, return a bytes object.

    Support multiple concatenated frames.

    Parameters
    data:      A bytes-like object, compressed zstd data.
    zstd_dict: A ZstdDict object, pre-trained zstd dictionary.
    option:    A dict object, contains advanced decompression parameters.
    """
    # EndlessZstdDecompressor
    decomp = EndlessZstdDecompressor(zstd_dict, option)

    # Prepare input data
    in_buf = decomp._singleton_in_buf
    in_buf.src = ffi.from_buffer(data)
    in_buf.size = len(data)
    in_buf.pos = 0

    # Get decompressed size
    decompressed_size = m.ZSTD_getFrameContentSize(ffi.from_buffer(data), len(data))
    if decompressed_size not in (m.ZSTD_CONTENTSIZE_UNKNOWN,
                                 m.ZSTD_CONTENTSIZE_ERROR):
        initial_size = decompressed_size
    else:
        initial_size = -1

    # Decompress
    ret = decomp._decompress_impl(in_buf, -1, initial_size)

    # Check data integrity. at_frame_edge flag is True when the both the input
    # and output streams are at a frame edge.
    if not decomp._at_frame_edge:
        extra_msg = "." if (len(ret) == 0) \
                        else (", if want to output these decompressed data, use "
                              "decompress_stream function or "
                              "EndlessZstdDecompressor class to decompress.")
        msg = ("Decompression failed: zstd data ends in an incomplete "
               "frame, maybe the input data was truncated. Decompressed "
               "data is %d bytes%s") % (len(ret), extra_msg)
        raise ZstdError(msg)

    return ret



from .common import ZstdError, CParameter, DParameter, Strategy, \
                    get_frame_info, get_frame_size, \
                    zstd_version, zstd_version_info, \
                    compressionLevel_values, \
                    _train_dict, _finalize_dict, \
                    _ZSTD_CStreamSizes, _ZSTD_DStreamSizes, \
                    PYZSTD_CONFIG
from .dict import ZstdDict
from .compressor import ZstdCompressor, RichMemZstdCompressor
from .decompressor import ZstdDecompressor, EndlessZstdDecompressor, decompress
from .stream import compress_stream, decompress_stream
from .file import ZstdFileReader, ZstdFileWriter

__all__ = ('ZstdCompressor', 'RichMemZstdCompressor',
           'ZstdDecompressor', 'EndlessZstdDecompressor',
           'ZstdDict', 'ZstdError',
           'CParameter', 'DParameter', 'Strategy',
           'decompress', 'get_frame_info', 'get_frame_size',
           'compress_stream', 'decompress_stream',
           'zstd_version', 'zstd_version_info',
           'compressionLevel_values',
           '_train_dict', '_finalize_dict',
           'ZstdFileReader', 'ZstdFileWriter',
           '_ZSTD_CStreamSizes', '_ZSTD_DStreamSizes',
           'PYZSTD_CONFIG')



import sys
from collections import namedtuple
from enum import IntEnum
from functools import lru_cache

from ._cffi_zstd import ffi, lib as m

PYZSTD_CONFIG = (64 if sys.maxsize > 2**32 else 32,
                 'cffi', bool(m.pyzstd_static_link), False)

_ZSTD_CStreamSizes = (m.ZSTD_CStreamInSize(), m.ZSTD_CStreamOutSize())
_ZSTD_DStreamSizes = (m.ZSTD_DStreamInSize(), m.ZSTD_DStreamOutSize())

zstd_version = ffi.string(m.ZSTD_versionString()).decode('ascii')
zstd_version_info = tuple(int(i) for i in zstd_version.split('.'))

_nt_values = namedtuple('values', ['default', 'min', 'max'])
compressionLevel_values = _nt_values(m.ZSTD_defaultCLevel(),
                                     m.ZSTD_minCLevel(),
                                     m.ZSTD_maxCLevel())

_new_nonzero = ffi.new_allocator(should_clear_after_alloc=False)

def _nbytes(dat):
    if isinstance(dat, (bytes, bytearray)):
        return len(dat)
    with memoryview(dat) as mv:
        return mv.nbytes

class ZstdError(Exception):
    "Call to the underlying zstd library failed."
    pass

def _get_param_bounds(is_compress, key):
    # Get parameter bounds
    if is_compress:
        bounds = m.ZSTD_cParam_getBounds(key)
        if m.ZSTD_isError(bounds.error):
            _set_zstd_error(_ErrorType.ERR_GET_C_BOUNDS, bounds.error)
    else:
        bounds = m.ZSTD_dParam_getBounds(key)
        if m.ZSTD_isError(bounds.error):
            _set_zstd_error(_ErrorType.ERR_GET_D_BOUNDS, bounds.error)

    return (bounds.lowerBound, bounds.upperBound)

class CParameter(IntEnum):
    """Compression parameters"""

    compressionLevel           = m.ZSTD_c_compressionLevel
    windowLog                  = m.ZSTD_c_windowLog
    hashLog                    = m.ZSTD_c_hashLog
    chainLog                   = m.ZSTD_c_chainLog
    searchLog                  = m.ZSTD_c_searchLog
    minMatch                   = m.ZSTD_c_minMatch
    targetLength               = m.ZSTD_c_targetLength
    strategy                   = m.ZSTD_c_strategy

    enableLongDistanceMatching = m.ZSTD_c_enableLongDistanceMatching
    ldmHashLog                 = m.ZSTD_c_ldmHashLog
    ldmMinMatch                = m.ZSTD_c_ldmMinMatch
    ldmBucketSizeLog           = m.ZSTD_c_ldmBucketSizeLog
    ldmHashRateLog             = m.ZSTD_c_ldmHashRateLog

    contentSizeFlag            = m.ZSTD_c_contentSizeFlag
    checksumFlag               = m.ZSTD_c_checksumFlag
    dictIDFlag                 = m.ZSTD_c_dictIDFlag

    nbWorkers                  = m.ZSTD_c_nbWorkers
    jobSize                    = m.ZSTD_c_jobSize
    overlapLog                 = m.ZSTD_c_overlapLog

    @lru_cache(maxsize=None)
    def bounds(self):
        """Return lower and upper bounds of a compression parameter, both inclusive."""
        # 1 means compression parameter
        return _get_param_bounds(1, self.value)

class DParameter(IntEnum):
    """Decompression parameters"""

    windowLogMax = m.ZSTD_d_windowLogMax

    @lru_cache(maxsize=None)
    def bounds(self):
        """Return lower and upper bounds of a decompression parameter, both inclusive."""
        # 0 means decompression parameter
        return _get_param_bounds(0, self.value)

class Strategy(IntEnum):
    """Compression strategies, listed from fastest to strongest.

    Note : new strategies _might_ be added in the future, only the order
    (from fast to strong) is guaranteed.
    """
    fast     = m.ZSTD_fast
    dfast    = m.ZSTD_dfast
    greedy   = m.ZSTD_greedy
    lazy     = m.ZSTD_lazy
    lazy2    = m.ZSTD_lazy2
    btlazy2  = m.ZSTD_btlazy2
    btopt    = m.ZSTD_btopt
    btultra  = m.ZSTD_btultra
    btultra2 = m.ZSTD_btultra2

class _ErrorType:
    ERR_DECOMPRESS=0
    ERR_COMPRESS=1
    ERR_SET_PLEDGED_INPUT_SIZE=2

    ERR_LOAD_D_DICT=3
    ERR_LOAD_C_DICT=4

    ERR_GET_C_BOUNDS=5
    ERR_GET_D_BOUNDS=6
    ERR_SET_C_LEVEL=7

    ERR_TRAIN_DICT=8
    ERR_FINALIZE_DICT=9

    _TYPE_MSG = (
        "Unable to decompress zstd data: %s",
        "Unable to compress zstd data: %s",
        "Unable to set pledged uncompressed content size: %s",

        "Unable to load zstd dictionary or prefix for decompression: %s",
        "Unable to load zstd dictionary or prefix for compression: %s",

        "Unable to get zstd compression parameter bounds: %s",
        "Unable to get zstd decompression parameter bounds: %s",
        "Unable to set zstd compression level: %s",

        "Unable to train zstd dictionary: %s",
        "Unable to finalize zstd dictionary: %s")

    @staticmethod
    def get_type_msg(type):
        return _ErrorType._TYPE_MSG[type]

def _set_zstd_error(type, zstd_ret):
    msg = _ErrorType.get_type_msg(type) % \
          ffi.string(m.ZSTD_getErrorName(zstd_ret)).decode('utf-8')
    raise ZstdError(msg)

def _set_parameter_error(is_compress, key, value):
    COMPRESS_PARAMETERS = \
    {m.ZSTD_c_compressionLevel: "compressionLevel",
     m.ZSTD_c_windowLog:        "windowLog",
     m.ZSTD_c_hashLog:          "hashLog",
     m.ZSTD_c_chainLog:         "chainLog",
     m.ZSTD_c_searchLog:        "searchLog",
     m.ZSTD_c_minMatch:         "minMatch",
     m.ZSTD_c_targetLength:     "targetLength",
     m.ZSTD_c_strategy:         "strategy",

     m.ZSTD_c_enableLongDistanceMatching: "enableLongDistanceMatching",
     m.ZSTD_c_ldmHashLog:       "ldmHashLog",
     m.ZSTD_c_ldmMinMatch:      "ldmMinMatch",
     m.ZSTD_c_ldmBucketSizeLog: "ldmBucketSizeLog",
     m.ZSTD_c_ldmHashRateLog:   "ldmHashRateLog",

     m.ZSTD_c_contentSizeFlag:  "contentSizeFlag",
     m.ZSTD_c_checksumFlag:     "checksumFlag",
     m.ZSTD_c_dictIDFlag:       "dictIDFlag",

     m.ZSTD_c_nbWorkers:        "nbWorkers",
     m.ZSTD_c_jobSize:          "jobSize",
     m.ZSTD_c_overlapLog:       "overlapLog"}

    DECOMPRESS_PARAMETERS = {m.ZSTD_d_windowLogMax: "windowLogMax"}

    if is_compress:
        parameters = COMPRESS_PARAMETERS
        type_msg = "compression"
    else:
        parameters = DECOMPRESS_PARAMETERS
        type_msg = "decompression"

    # Find parameter's name
    name = parameters.get(key)
    # Unknown parameter
    if name is None:
        name = 'unknown parameter (key %d)' % key

    # Get parameter bounds
    if is_compress:
        bounds = m.ZSTD_cParam_getBounds(key)
    else:
        bounds = m.ZSTD_dParam_getBounds(key)
    if m.ZSTD_isError(bounds.error):
        msg = 'Zstd %s parameter "%s" is invalid. (zstd v%s)' % \
              (type_msg, name, zstd_version)
        raise ZstdError(msg)

    # Error message
    msg = ('Error when setting zstd %s parameter "%s", it '
           'should %d <= value <= %d, provided value is %d. '
           '(zstd v%s, %d-bit build)') % \
          (type_msg, name,
           bounds.lowerBound, bounds.upperBound, value,
           zstd_version, PYZSTD_CONFIG[0])
    raise ZstdError(msg)

def _check_int32_value(value, name):
    try:
        if value > 2147483647 or value < -2147483648:
            raise Exception
    except:
        raise ValueError("%s should be 32-bit signed int value." % name)

# return: (compressionLevel, use_multithread)
def _set_c_parameters(cctx, level_or_option):
    if isinstance(level_or_option, int):
        _check_int32_value(level_or_option, "Compression level")

        # Set compression level
        zstd_ret = m.ZSTD_CCtx_setParameter(cctx, m.ZSTD_c_compressionLevel,
                                            level_or_option)
        if m.ZSTD_isError(zstd_ret):
            _set_zstd_error(_ErrorType.ERR_SET_C_LEVEL, zstd_ret)

        return level_or_option, False

    if isinstance(level_or_option, dict):
        level = 0  # 0 means use zstd's default compression level
        use_multithread = False

        for key, value in level_or_option.items():
            # Check key type
            if type(key) == DParameter:
                raise TypeError("Key of compression option dict should "
                                "NOT be DParameter.")

            # Both key & value should be 32-bit signed int
            _check_int32_value(key, "Key of option dict")
            _check_int32_value(value, "Value of option dict")

            if key == m.ZSTD_c_compressionLevel:
                level = value
            elif key == m.ZSTD_c_nbWorkers:
                if value != 0:
                    use_multithread = True

            # Set parameter
            zstd_ret = m.ZSTD_CCtx_setParameter(cctx, key, value)
            if m.ZSTD_isError(zstd_ret):
                _set_parameter_error(True, key, value)

        return level, use_multithread

    raise TypeError("level_or_option argument wrong type.")

def _set_d_parameters(dctx, option):
    if not isinstance(option, dict):
        raise TypeError("option argument should be dict object.")

    for key, value in option.items():
        # Check key type
        if type(key) == CParameter:
            raise TypeError("Key of decompression option dict should "
                            "NOT be CParameter.")

        # Both key & value should be 32-bit signed int
        _check_int32_value(key, "Key of option dict")
        _check_int32_value(value, "Value of option dict")

        # Set parameter
        zstd_ret = m.ZSTD_DCtx_setParameter(dctx, key, value)
        if m.ZSTD_isError(zstd_ret):
            _set_parameter_error(False, key, value)

# Write output data to fp.
# If (out_b.pos == 0), do nothing.
def _write_to_fp(func_name, fp, out_mv, out_b):
    if out_b.pos == 0:
        return

    write_ret = fp.write(out_mv[:out_b.pos])
    if write_ret != out_b.pos:
        msg = ("%s returned invalid length %d "
               "(should be %d <= value <= %d)") % \
               (func_name, write_ret, out_b.pos, out_b.pos)
        raise ValueError(msg)

def _train_dict(samples_bytes, samples_size_list, dict_size):
    # C code
    if dict_size <= 0:
        raise ValueError("dict_size argument should be positive number.")

    # Prepare chunk_sizes
    _chunks_number = len(samples_size_list)
    _sizes = _new_nonzero("size_t[]", _chunks_number)
    if _sizes == ffi.NULL:
        raise MemoryError

    _sizes_sum = 0
    for i, size in enumerate(samples_size_list):
        _sizes[i] = size
        _sizes_sum += size

    if _sizes_sum != _nbytes(samples_bytes):
        msg = "The samples size list doesn't match the concatenation's size."
        raise ValueError(msg)

    # Allocate dict buffer
    _dst_dict_bytes = _new_nonzero("char[]", dict_size)
    if _dst_dict_bytes == ffi.NULL:
        raise MemoryError

    # Train
    zstd_ret = m.ZDICT_trainFromBuffer(_dst_dict_bytes, dict_size,
                                       ffi.from_buffer(samples_bytes),
                                       _sizes, _chunks_number)
    if m.ZDICT_isError(zstd_ret):
        _set_zstd_error(_ErrorType.ERR_TRAIN_DICT, zstd_ret)

    # Resize dict_buffer
    b = ffi.buffer(_dst_dict_bytes)[:zstd_ret]
    return b

def _finalize_dict(custom_dict_bytes,
                   samples_bytes, samples_size_list,
                   dict_size, compression_level):
    # If m.ZSTD_VERSION_NUMBER < 10405, m.ZDICT_finalizeDictionary() is an
    # empty function defined in build_cffi.py.
    # If m.ZSTD_versionNumber() < 10405, m.ZDICT_finalizeDictionary() doesn't
    # exist in run-time zstd library.
    if (m.ZSTD_VERSION_NUMBER < 10405          # compile-time version
          or m.ZSTD_versionNumber() < 10405):  # run-time version
        msg = ("finalize_dict function only available when the underlying "
               "zstd library's version is greater than or equal to v1.4.5. "
               "At pyzstd module's compile-time, zstd version is %d. At "
               "pyzstd module's run-time, zstd version is %d.") % \
               (m.ZSTD_VERSION_NUMBER, m.ZSTD_versionNumber())
        raise NotImplementedError(msg)

    # C code
    if dict_size <= 0:
        raise ValueError("dict_size argument should be positive number.")

    # Prepare chunk_sizes
    _chunks_number = len(samples_size_list)
    _sizes = _new_nonzero("size_t[]", _chunks_number)
    if _sizes == ffi.NULL:
        raise MemoryError

    _sizes_sum = 0
    for i, size in enumerate(samples_size_list):
        _sizes[i] = size
        _sizes_sum += size

    if _sizes_sum != _nbytes(samples_bytes):
        msg = "The samples size list doesn't match the concatenation's size."
        raise ValueError(msg)

    # Allocate dict buffer
    _dst_dict_bytes = _new_nonzero("char[]", dict_size)
    if _dst_dict_bytes == ffi.NULL:
        raise MemoryError

    # Parameters
    params = _new_nonzero("ZDICT_params_t *")
    if params == ffi.NULL:
        raise MemoryError
    # Optimize for a specific zstd compression level, 0 means default.
    params.compressionLevel = compression_level
    # Write log to stderr, 0 = none.
    params.notificationLevel = 0
    # Force dictID value, 0 means auto mode (32-bits random value).
    params.dictID = 0

    # Finalize
    zstd_ret = m.ZDICT_finalizeDictionary(
                   _dst_dict_bytes, dict_size,
                   ffi.from_buffer(custom_dict_bytes), _nbytes(custom_dict_bytes),
                   ffi.from_buffer(samples_bytes), _sizes, _chunks_number,
                   params[0])
    if m.ZDICT_isError(zstd_ret):
        _set_zstd_error(_ErrorType.ERR_FINALIZE_DICT, zstd_ret)

    # Resize dict_buffer
    b = ffi.buffer(_dst_dict_bytes)[:zstd_ret]
    return b

_nt_frame_info = namedtuple('frame_info',
                            ['decompressed_size', 'dictionary_id'])

def get_frame_info(frame_buffer):
    """Get zstd frame information from a frame header.

    Parameter
    frame_buffer: A bytes-like object. It should starts from the beginning of
                  a frame, and needs to include at least the frame header (6 to
                  18 bytes).

    Return a two-items namedtuple: (decompressed_size, dictionary_id)

    If decompressed_size is None, decompressed size is unknown.

    dictionary_id is a 32-bit unsigned integer value. 0 means dictionary ID was
    not recorded in the frame header, the frame may or may not need a dictionary
    to be decoded, and the ID of such a dictionary is not specified.

    It's possible to append more items to the namedtuple in the future.
    """

    decompressed_size = m.ZSTD_getFrameContentSize(
                            ffi.from_buffer(frame_buffer), len(frame_buffer))
    if decompressed_size == m.ZSTD_CONTENTSIZE_UNKNOWN:
        decompressed_size = None
    elif decompressed_size == m.ZSTD_CONTENTSIZE_ERROR:
        msg = ("Error when getting information from the header of "
               "a zstd frame. Make sure the frame_buffer argument "
               "starts from the beginning of a frame, and its length "
               "not less than the frame header (6~18 bytes).")
        raise ZstdError(msg)

    dict_id = m.ZSTD_getDictID_fromFrame(
                  ffi.from_buffer(frame_buffer), len(frame_buffer))

    ret = _nt_frame_info(decompressed_size, dict_id)
    return ret

def get_frame_size(frame_buffer):
    """Get the size of a zstd frame, including frame header and 4-byte checksum if it
    has.

    It will iterate all blocks' header within a frame, to accumulate the frame size.

    Parameter
    frame_buffer: A bytes-like object, it should starts from the beginning of a
                  frame, and contains at least one complete frame.
    """

    frame_size = m.ZSTD_findFrameCompressedSize(
                     ffi.from_buffer(frame_buffer), len(frame_buffer))
    if m.ZSTD_isError(frame_size):
        msg = ("Error when finding the compressed size of a zstd frame. "
               "Make sure the frame_buffer argument starts from the "
               "beginning of a frame, and its length not less than this "
               "complete frame. Zstd error message: %s.") % \
               ffi.string(m.ZSTD_getErrorName(frame_size)).decode('utf-8')
        raise ZstdError(msg)

    return frame_size



#!/usr/bin/env python3
# CLI of pyzstd module: python -m pyzstd --help
import argparse
import os
from time import time

from pyzstd import compress_stream, decompress_stream, \
                   CParameter, DParameter, \
                   train_dict, ZstdDict, ZstdFile, \
                   compressionLevel_values, zstd_version, \
                   __version__ as pyzstd_version, PYZSTD_CONFIG

# buffer sizes recommended by zstd
C_READ_BUFFER = 131072
D_READ_BUFFER = 131075

def progress_bar(progress, total, width=50):
    # documented behavior: if input stream is empty, the callback
    # function will not be called. so no ZeroDivisionError here.
    percent = (progress + 1) / total
    now = int(width * percent)
    bar = "+" * now + "-" * (width - now)
    print("|%s| %.2f%%" % (bar, 100*percent), end="\r", flush=True)

# open output file and assign to args.output
def open_output(args, path):
    if not args.f and os.path.isfile(path):
        answer = input(('output file already exists:\n'
                        '{}\noverwrite? (y/n) ').format(path))
        print()
        if answer != 'y':
            import sys
            sys.exit()
    args.output = open(path, 'wb')

def close_files(args):
    if args.input is not None:
        args.input.close()

    if args.output is not None:
        args.output.close()

def compress_option(args):
    # threads message
    if args.threads == 0:
        threads_msg = 'single-thread mode'
    else:
        threads_msg = 'multi-thread mode, %d threads.' % args.threads

    # long mode
    if args.long >= 0:
        use_long = 1
        windowLog = args.long
        long_msg = 'yes, windowLog is %d.' % windowLog
    else:
        use_long = 0
        windowLog = 0
        long_msg = 'no'

    # option
    option = {CParameter.compressionLevel: args.level,
              CParameter.nbWorkers: args.threads,
              CParameter.enableLongDistanceMatching: use_long,
              CParameter.windowLog: windowLog,
              CParameter.checksumFlag: args.checksum,
              CParameter.dictIDFlag: args.write_dictID}

    # pre-compress message
    msg = (' - compression level: {}\n'
           ' - threads: {}\n'
           ' - long mode: {}\n'
           ' - zstd dictionary: {}\n'
           ' - add checksum: {}').format(
                args.level, threads_msg, long_msg,
                args.zd, args.checksum)
    print(msg)

    return option

def compress(args):
    # input file size
    input_file_size = os.path.getsize(args.input.name)

    # output file
    if args.output is None:
        open_output(args, args.input.name + '.zst')

    # pre-compress message
    msg = ('Compress file:\n'
           ' - input file : {}\n'
           ' - output file: {}').format(
                args.input.name,
                args.output.name)
    print(msg)

    # option
    option = compress_option(args)

    # compress
    def cb(total_input, total_output, read_data, write_data):
        progress_bar(total_input, input_file_size)
    t1 = time()
    ret = compress_stream(args.input, args.output,
                          level_or_option=option,
                          zstd_dict=args.zd,
                          pledged_input_size=input_file_size,
                          callback=cb)
    t2 = time()
    close_files(args)

    # post-compress message
    if ret[0] != 0:
        ratio = 100 * ret[1] / ret[0]
    else:
        ratio = 100.0
    msg = ('\nCompression succeeded, {:.2f} seconds.\n'
           'Input {:,} bytes, output {:,} bytes, ratio {:.2f}%.\n').format(
            t2-t1, *ret, ratio)
    print(msg)

def decompress(args):
    # input file size
    input_file_size = os.path.getsize(args.input.name)

    # output file
    if args.output is None and args.test is None:
        from re import subn

        out_path, replaced = subn(r'(?i)^(.*)\.zst$', r'\1', args.input.name)
        if not replaced:
            out_path = args.input.name + '.decompressed'
        open_output(args, out_path)

    # option
    option = {DParameter.windowLogMax: args.windowLogMax}

    # pre-decompress message
    if args.output is not None:
        output_name = args.output.name
    else:
        output_name = 'None'
    print(('Decompress file:\n'
           ' - input file : {}\n'
           ' - output file: {}\n'
           ' - zstd dictionary: {}').format(
            args.input.name, output_name, args.zd))

    # decompress
    def cb(total_input, total_output, read_data, write_data):
        progress_bar(total_input, input_file_size)
    t1 = time()
    ret = decompress_stream(args.input, args.output,
                            zstd_dict=args.zd, option=option,
                            callback=cb)
    t2 = time()
    close_files(args)

    # post-decompress message
    if ret[1] != 0:
        ratio = 100 * ret[0] / ret[1]
    else:
        ratio = 100.0
    msg = ('\nDecompression succeeded, {:.2f} seconds.\n'
           'Input {:,} bytes, output {:,} bytes, ratio {:.2f}%.\n').format(
            t2-t1, *ret, ratio)
    print(msg)

def train(args):
    from glob import glob

    # check output file
    if args.output is None:
        msg = 'need to specify output file using -o/--output option'
        raise Exception(msg)

    # gather samples
    print('Gathering samples, please wait.', flush=True)
    lst = []
    for file in glob(args.train, recursive=True):
        with open(file, 'rb') as f:
            dat = f.read()
            lst.append(dat)
            print('samples count: %d' % len(lst), end='\r', flush=True)
    if len(lst) == 0:
        raise Exception('No samples gathered, please check GLOB_PATH.')

    samples_size = sum(len(sample) for sample in lst)
    if samples_size == 0:
        raise Exception("Samples content is empty, can't train.")

    # pre-train message
    msg = ('Gathered, train zstd dictionary:\n'
           ' - samples: {}\n'
           ' - samples number: {}\n'
           ' - samples content: {:,} bytes\n'
           ' - dict file: {}\n'
           ' - dict max size: {:,} bytes\n'
           ' - dict id: {}\n'
           'Training, please wait.').format(
                args.train, len(lst), samples_size,
                args.output.name, args.maxdict,
                'random' if args.dictID is None else args.dictID)
    print(msg, flush=True)

    # train
    t1 = time()
    zd = train_dict(lst, args.maxdict)
    t2 = time()

    # Dictionary_ID: 4 bytes, stored in little-endian format.
    # it can be any value, except 0 (which means no Dictionary_ID).
    if args.dictID is not None and len(zd.dict_content) >= 8:
        content = zd.dict_content[:4] + \
                    args.dictID.to_bytes(4, 'little') + \
                    zd.dict_content[8:]
        zd = ZstdDict(content)

    # save to file
    args.output.write(zd.dict_content)
    close_files(args)

    # post-train message
    msg = ('Training succeeded, {:.2f} seconds.\n'
           'Dictionary: {}\n').format(t2-t1, zd)
    print(msg)

def get_ZstdTarFile():
    # lazy import for tar operations
    from tarfile import TarFile

    class ZstdTarFile(TarFile):
        def __init__(self, name, mode='r', *, level_or_option=None, zstd_dict=None, **kwargs):
            self.zstd_file = ZstdFile(name, mode,
                                      level_or_option=level_or_option,
                                      zstd_dict=zstd_dict)
            try:
                super().__init__(fileobj=self.zstd_file, mode=mode, **kwargs)
            except:
                self.zstd_file.close()
                raise

        def close(self):
            try:
                super().close()
            finally:
                self.zstd_file.close()

    return ZstdTarFile

def tarfile_create(args):
    # check input dir
    args.tar_input_dir = args.tar_input_dir.rstrip(os.sep)
    if not os.path.isdir(args.tar_input_dir):
        msg = 'Tar archive input dir invalid: ' + args.tar_input_dir
        raise NotADirectoryError(msg)
    dirname, basename = os.path.split(args.tar_input_dir)

    # check output file
    if args.output is None:
        out_path = os.path.join(dirname, basename + '.tar.zst')
        open_output(args, out_path)

    # get ZstdTarFile class
    ZstdTarFile = get_ZstdTarFile()

    # pre-compress message
    msg = ('Archive tar file:\n'
           ' - input directory: {}\n'
           ' - output file: {}').format(
                args.tar_input_dir,
                args.output.name)
    print(msg)

    # option
    option = compress_option(args)

    # compress
    print('Archiving, please wait.', flush=True)
    t1 = time()
    with ZstdTarFile(args.output, mode='w',
                     level_or_option=option,
                     zstd_dict=args.zd) as f:
        f.add(args.tar_input_dir, basename)
        uncompressed_size = f.fileobj.tell()
    t2 = time()

    output_file_size = args.output.tell()
    close_files(args)

    # post-compress message
    if uncompressed_size != 0:
        ratio = 100 * output_file_size / uncompressed_size
    else:
        ratio = 100.0

    msg = ('Archiving succeeded, {:.2f} seconds.\n'
           'Input ~{:,} bytes, output {:,} bytes, ratio {:.2f}%.\n').format(
            t2-t1, uncompressed_size, output_file_size, ratio)
    print(msg)

def tarfile_extract(args):
    # input file size
    if args.input is None:
        msg = 'need to specify input file using -d/--decompress option.'
        raise FileNotFoundError(msg)
    input_file_size = os.path.getsize(args.input.name)

    # check output dir
    if not os.path.isdir(args.tar_output_dir):
        msg = 'Tar archive output dir invalid: ' + args.tar_output_dir
        raise NotADirectoryError(msg)

    # get ZstdTarFile class
    ZstdTarFile = get_ZstdTarFile()

    # option
    option = {DParameter.windowLogMax: args.windowLogMax}

    # pre-extract message
    msg = ('Extract tar archive:\n'
           ' - input file: {}\n'
           ' - output dir: {}\n'
           ' - zstd dictionary: {}\n'
           'Extracting, please wait.').format(
                args.input.name,
                args.tar_output_dir,
                args.zd)
    print(msg, flush=True)

    # extract
    t1 = time()
    with ZstdTarFile(args.input, mode='r',
                     zstd_dict=args.zd,
                     level_or_option=option) as f:
        f.extractall(args.tar_output_dir)
        uncompressed_size = f.fileobj.tell()
    t2 = time()
    close_files(args)

    # post-extract message
    if uncompressed_size != 0:
        ratio = 100 * input_file_size / uncompressed_size
    else:
        ratio = 100.0
    msg = ('Extraction succeeded, {:.2f} seconds.\n'
           'Input {:,} bytes, output ~{:,} bytes, ratio {:.2f}%.\n').format(
            t2-t1, input_file_size, uncompressed_size, ratio)
    print(msg)

def range_action(min, max, bits_msg=False):
    class RangeAction(argparse.Action):
        def __call__(self, parser, args, values, option_string=None):
            # convert to int
            try:
                v = int(values)
            except:
                raise TypeError('{} should be an integer'.format(option_string))

            # check range
            if not (min <= v <= max):
                # 32/64 bits message
                if bits_msg:
                    bits = 'in {}-bit build, '.format(PYZSTD_CONFIG[0])
                else:
                    bits = ''

                # message
                msg = ('{}{} value should: {} <= v <= {}. '
                       'provided value is {}.').format(
                        bits, option_string, min, max, v)
                raise ValueError(msg)

            setattr(args, self.dest, v)
    return RangeAction

def parse_arg():
    p = argparse.ArgumentParser(
                    prog = 'CLI of pyzstd module',
                    description = ("The command style is similar to zstd's "
                                   "CLI, but there are some differences.\n"
                                   "Zstd's CLI should be faster, it has "
                                   "some I/O optimizations."),
                    epilog=('Examples of use:\n'
                            '  compress a file:\n'
                            '    python -m pyzstd -c IN_FILE -o OUT_FILE\n'
                            '  decompress a file:\n'
                            '    python -m pyzstd -d IN_FILE -o OUT_FILE\n'
                            '  create a tar archive:\n'
                            '    python -m pyzstd --tar-input-dir DIR -o OUT_FILE\n'
                            '  extract a tar archive, output will forcibly overwrite existing files:\n'
                            '    python -m pyzstd -d IN_FILE --tar-output-dir DIR\n'
                            '  train a zstd dictionary, ** traverses sub-directories:\n'
                            '    python -m pyzstd --train "E:\\cpython\\**\\*.c" -o OUT_FILE'),
                    formatter_class=argparse.RawDescriptionHelpFormatter)

    g = p.add_argument_group('Common arguments')
    g.add_argument('-D', '--dict', metavar='FILE', type=argparse.FileType('rb'),
                   help='use FILE as zstd dictionary for compression or decompression')
    g.add_argument('-o', '--output', metavar='FILE', type=str,
                   help='result stored into FILE')
    g.add_argument('-f', action='store_true',
                   help='disable output check, allows overwriting existing file.')

    g = p.add_argument_group('Compression arguments')
    gm = g.add_mutually_exclusive_group()
    gm.add_argument('-c', '--compress', metavar='FILE', type=str,
                    help='compress FILE')
    gm.add_argument('--tar-input-dir', metavar='DIR', type=str,
                    help=('create a tar archive from DIR. '
                          'this option overrides -c/--compress option.'))
    g.add_argument('-l', '--level', metavar='#',
                   default=compressionLevel_values.default,
                   action=range_action(compressionLevel_values.min,
                                       compressionLevel_values.max),
                   help='compression level, range: [{},{}], default: {}.'.
                   format(compressionLevel_values.min,
                          compressionLevel_values.max,
                          compressionLevel_values.default))
    g.add_argument('-t', '--threads', metavar='#', default=0,
                   action=range_action(*CParameter.nbWorkers.bounds(), True),
                   help=('spawns # threads to compress. if this option is not '
                         'specified or is 0, use single thread mode.'))
    g.add_argument('--long', metavar='#', nargs='?', const=27, default=-1,
                   action=range_action(*CParameter.windowLog.bounds(), True),
                   help='enable long distance matching with given windowLog (default #: 27)')
    g.add_argument('--no-checksum', action='store_false',
                   dest='checksum', default=True,
                   help="don't add 4-byte XXH64 checksum to the frame")
    g.add_argument('--no-dictID', action='store_false',
                   dest='write_dictID', default=True,
                   help="don't write dictID into frame header (dictionary compression only)")

    g = p.add_argument_group('Decompression arguments')
    gm = g.add_mutually_exclusive_group()
    gm.add_argument('-d', '--decompress', metavar='FILE', type=str,
                    help='decompress FILE')
    g.add_argument('--tar-output-dir', metavar='DIR', type=str,
                   help=('extract tar archive to DIR, '
                         'output will forcibly overwrite existing files. '
                         'this option overrides -o/--output option.'))
    gm.add_argument('--test', metavar='FILE', type=str,
                    help='try to decompress FILE to check integrity')
    g.add_argument('--windowLogMax', metavar='#', default=0,
                   action=range_action(*DParameter.windowLogMax.bounds(), True),
                   help='set a memory usage limit for decompression (windowLogMax)')

    g = p.add_argument_group('Dictionary builder')
    g.add_argument('--train', metavar='GLOB_PATH', type=str,
                   help='create a dictionary from a training set of files')
    g.add_argument('--maxdict', metavar='SIZE', type=int, default=112640,
                   help='limit dictionary to SIZE bytes (default: 112640)')
    g.add_argument('--dictID', metavar='DICT_ID', default=None,
                   action=range_action(1, 0xFFFFFFFF),
                   help='specify dictionary ID value (default: random)')

    args = p.parse_args()

    # input file
    if args.compress is not None:
        args.input = open(args.compress, 'rb',
                          buffering=C_READ_BUFFER)
    elif args.decompress is not None:
        args.input = open(args.decompress, 'rb',
                          buffering=D_READ_BUFFER)
    elif args.test is not None:
        args.input = open(args.test, 'rb',
                          buffering=D_READ_BUFFER)
    else:
        args.input = None

    # output file
    if args.output is not None:
        open_output(args, args.output)

    # load dictionary
    if args.dict is not None:
        zd_content = args.dict.read()
        args.dict.close()
        # Magic_Number: 4 bytes, value 0xEC30A437, little-endian format.
        is_raw = zd_content[:4] != b'\x37\xA4\x30\xEC'
        args.zd = ZstdDict(zd_content, is_raw)
    else:
        args.zd = None

    # arguments combination
    functions = [args.compress, args.decompress,
                 args.test, args.train, args.tar_input_dir]
    if sum(1 for i in functions if i is not None) > 1:
        raise Exception('Wrong arguments combination')

    return args

def main():
    print('*** pyzstd module v{}, zstd library v{}. ***\n'.
          format(pyzstd_version, zstd_version))

    args = parse_arg()

    if args.tar_input_dir:
        tarfile_create(args)
    elif args.tar_output_dir:
        tarfile_extract(args)
    elif args.compress:
        compress(args)
    elif args.decompress or args.test:
        decompress(args)
    elif args.train:
        train(args)
    else:
        print('Invalid command. See help: python -m pyzstd --help')

if __name__ == '__main__':
    main()



from .common import m, ffi, _new_nonzero

class _BlocksOutputBuffer:
    KB = 1024
    MB = 1024 * 1024
    BUFFER_BLOCK_SIZE = (
        # If change this list, also change:
        #   The C implementation
        #   OutputBufferTestCase unittest
        # If change the first blocks's size, also change:
        #   _32_KiB in __init__.py
        #   FileTestCase.test_decompress_limited() test
        32*KB, 64*KB, 256*KB, 1*MB, 4*MB, 8*MB, 16*MB, 16*MB,
        32*MB, 32*MB, 32*MB, 32*MB, 64*MB, 64*MB, 128*MB, 128*MB,
        256*MB )
    MEM_ERR_MSG = "Unable to allocate output buffer."

    def initAndGrow(self, out, max_length):
        # Get block size
        if 0 <= max_length < self.BUFFER_BLOCK_SIZE[0]:
            block_size = max_length
        else:
            block_size = self.BUFFER_BLOCK_SIZE[0]

        # The first block
        block = _new_nonzero("char[]", block_size)
        if block == ffi.NULL:
            raise MemoryError

        # Create the list
        self.list = [block]

        # Set variables
        self.allocated = block_size
        self.max_length = max_length

        out.dst = block
        out.size = block_size
        out.pos = 0

    def initWithSize(self, out, max_length, init_size):
        # Get block size
        if 0 <= max_length < init_size:
            block_size = max_length
        else:
            block_size = init_size

        # The first block
        block = _new_nonzero("char[]", block_size)
        if block == ffi.NULL:
            raise MemoryError(self.MEM_ERR_MSG)

        # Create the list
        self.list = [block]

        # Set variables
        self.allocated = block_size
        self.max_length = max_length

        out.dst = block
        out.size = block_size
        out.pos = 0

    def grow(self, out):
        # Ensure no gaps in the data
        assert out.pos == out.size

        # Get block size
        list_len = len(self.list)
        if list_len < len(self.BUFFER_BLOCK_SIZE):
            block_size = self.BUFFER_BLOCK_SIZE[list_len]
        else:
            block_size = self.BUFFER_BLOCK_SIZE[-1]

        # Check max_length
        if self.max_length >= 0:
            # If (rest == 0), should not grow the buffer.
            rest = self.max_length - self.allocated
            assert rest > 0

            # block_size of the last block
            if block_size > rest:
                block_size = rest

        # Create the block
        b = _new_nonzero("char[]", block_size)
        if b == ffi.NULL:
            raise MemoryError(self.MEM_ERR_MSG)
        self.list.append(b)

        # Set variables
        self.allocated += block_size

        out.dst = b
        out.size = block_size
        out.pos = 0

    def reachedMaxLength(self, out):
        # Ensure (data size == allocated size)
        assert out.pos == out.size

        return self.allocated == self.max_length

    def finish(self, out):
        # Fast path for single block
        if (len(self.list) == 1 and out.pos == out.size) or \
           (len(self.list) == 2 and out.pos == 0):
            return bytes(ffi.buffer(self.list[0]))

        # Final bytes object
        data_size = self.allocated - (out.size-out.pos)
        final = _new_nonzero("char[]", data_size)
        if final == ffi.NULL:
            raise MemoryError(self.MEM_ERR_MSG)

        # Memory copy
        # Blocks except the last one
        posi = 0
        for block in self.list[:-1]:
            ffi.memmove(final+posi, block, len(block))
            posi += len(block)
        # The last block
        ffi.memmove(final+posi, self.list[-1], out.pos)

        return bytes(ffi.buffer(final))


from threading import Lock
from warnings import warn

from .common import m, ffi, _new_nonzero, _nbytes, \
                    ZstdError, _set_c_parameters, \
                    _set_zstd_error, _ErrorType
from .dict import _load_c_dict
from .output_buffer import _BlocksOutputBuffer

class _Compressor:
    def __init__(self, level_or_option=None, zstd_dict=None):
        self._use_multithread = False
        self._lock = Lock()
        level = 0  # 0 means use zstd's default compression level

        self._singleton_in_buf = _new_nonzero("ZSTD_inBuffer *")
        if self._singleton_in_buf == ffi.NULL:
            raise MemoryError

        self._singleton_out_buf = _new_nonzero("ZSTD_outBuffer *")
        if self._singleton_out_buf == ffi.NULL:
            raise MemoryError

        # Compression context
        self._cctx = m.ZSTD_createCCtx()
        if self._cctx == ffi.NULL:
            raise ZstdError("Unable to create ZSTD_CCtx instance.")

        # Set compressLevel/option to compression context
        if level_or_option is not None:
            level, self._use_multithread = \
                  _set_c_parameters(self._cctx, level_or_option)

        # Load dictionary to compression context
        if zstd_dict is not None:
            _load_c_dict(self._cctx, zstd_dict, level)
            self.__dict = zstd_dict

    def __del__(self):
        try:
            m.ZSTD_freeCCtx(self._cctx)
            self._cctx = ffi.NULL
        except AttributeError:
            pass

    def _compress_impl(self, data, end_directive, rich_mem):
        # Input buffer
        in_buf = self._singleton_in_buf
        in_buf.src = ffi.from_buffer(data)
        in_buf.size = _nbytes(data)
        in_buf.pos = 0

        # Output buffer
        out_buf = self._singleton_out_buf
        out = _BlocksOutputBuffer()

        # Initialize output buffer
        if rich_mem:
            init_size = m.ZSTD_compressBound(_nbytes(data))
            out.initWithSize(out_buf, -1, init_size)
        else:
            out.initAndGrow(out_buf, -1)

        while True:
            # Compress
            zstd_ret = m.ZSTD_compressStream2(self._cctx, out_buf, in_buf, end_directive)
            if m.ZSTD_isError(zstd_ret):
                _set_zstd_error(_ErrorType.ERR_COMPRESS, zstd_ret)

            # Finished
            if zstd_ret == 0:
                return out.finish(out_buf)

            # Output buffer should be exhausted, grow the buffer.
            if out_buf.pos == out_buf.size:
                out.grow(out_buf)

    def _compress_mt_continue_impl(self, data):
        # Input buffer
        in_buf = self._singleton_in_buf
        in_buf.src = ffi.from_buffer(data)
        in_buf.size = _nbytes(data)
        in_buf.pos = 0

        # Output buffer
        out_buf = self._singleton_out_buf
        out = _BlocksOutputBuffer()
        out.initAndGrow(out_buf, -1)

        while True:
            # Compress
            while True:
                zstd_ret = m.ZSTD_compressStream2(self._cctx,
                                                  out_buf, in_buf,
                                                  m.ZSTD_e_continue)
                if (out_buf.pos == out_buf.size
                      or in_buf.pos == in_buf.size
                      or m.ZSTD_isError(zstd_ret)):
                    break

            # Check error
            if m.ZSTD_isError(zstd_ret):
                _set_zstd_error(_ErrorType.ERR_COMPRESS, zstd_ret)

            # Like ._compress_impl(), output as much as possible.
            if out_buf.pos == out_buf.size:
                out.grow(out_buf)
            elif in_buf.pos == in_buf.size:
                # Finished
                return out.finish(out_buf)

    def __reduce__(self):
        msg = "Cannot pickle %s object." % type(self)
        raise TypeError(msg)

class ZstdCompressor(_Compressor):
    """A streaming compressor. Thread-safe at method level."""

    CONTINUE = m.ZSTD_e_continue
    """Used for mode parameter in .compress() method.

    Collect more data, encoder decides when to output compressed result, for optimal
    compression ratio. Usually used for traditional streaming compression.
    """

    FLUSH_BLOCK = m.ZSTD_e_flush
    """Used for mode parameter in .compress(), .flush() methods.

    Flush any remaining data, but don't close the current frame. Usually used for
    communication scenarios.

    If there is data, it creates at least one new block, that can be decoded
    immediately on reception. If no remaining data, no block is created, return b''.

    Note: Abuse of this mode will reduce compression ratio. Use it only when
    necessary.
    """

    FLUSH_FRAME = m.ZSTD_e_end
    """Used for mode parameter in .compress(), .flush() methods.

    Flush any remaining data, and close the current frame. Usually used for
    traditional flush.

    Since zstd data consists of one or more independent frames, data can still be
    provided after a frame is closed.

    Note: Abuse of this mode will reduce compression ratio, and some programs can
    only decompress single frame data. Use it only when necessary.
    """

    def __init__(self, level_or_option=None, zstd_dict=None):
        """Initialize a ZstdCompressor object.

        Parameters
        level_or_option: When it's an int object, it represents the compression level.
                         When it's a dict object, it contains advanced compression
                         parameters.
        zstd_dict:       A ZstdDict object, pre-trained zstd dictionary.
        """
        super().__init__(level_or_option=level_or_option, zstd_dict=zstd_dict)
        self.__last_mode = m.ZSTD_e_end

    def compress(self, data, mode=CONTINUE):
        """Provide data to the compressor object.
        Return a chunk of compressed data if possible, or b'' otherwise.

        Parameters
        data: A bytes-like object, data to be compressed.
        mode: Can be these 3 values .CONTINUE, .FLUSH_BLOCK, .FLUSH_FRAME.
        """
        if mode not in (ZstdCompressor.CONTINUE,
                        ZstdCompressor.FLUSH_BLOCK,
                        ZstdCompressor.FLUSH_FRAME):
            msg = ("mode argument wrong value, it should be one of "
                   "ZstdCompressor.CONTINUE, ZstdCompressor.FLUSH_BLOCK, "
                   "ZstdCompressor.FLUSH_FRAME.")
            raise ValueError(msg)

        with self._lock:
            try:
                if self._use_multithread and mode == ZstdCompressor.CONTINUE:
                    ret = self._compress_mt_continue_impl(data)
                else:
                    ret = self._compress_impl(data, mode, False)
                self.__last_mode = mode
                return ret
            except:
                self.__last_mode = m.ZSTD_e_end
                # Resetting cctx's session never fail
                m.ZSTD_CCtx_reset(self._cctx, m.ZSTD_reset_session_only)
                raise

    def flush(self, mode=FLUSH_FRAME):
        """Flush any remaining data in internal buffer.

        Since zstd data consists of one or more independent frames, the compressor
        object can still be used after this method is called.

        Parameter
        mode: Can be these 2 values .FLUSH_FRAME, .FLUSH_BLOCK.
        """
        if mode not in (ZstdCompressor.FLUSH_FRAME, ZstdCompressor.FLUSH_BLOCK):
            msg = ("mode argument wrong value, it should be "
                   "ZstdCompressor.FLUSH_FRAME or ZstdCompressor.FLUSH_BLOCK.")
            raise ValueError(msg)

        with self._lock:
            try:
                ret = self._compress_impl(b"", mode, False)
                self.__last_mode = mode
                return ret
            except:
                self.__last_mode = m.ZSTD_e_end
                # Resetting cctx's session never fail
                m.ZSTD_CCtx_reset(self._cctx, m.ZSTD_reset_session_only)
                raise

    def _set_pledged_input_size(self, size):
        """*This is an undocumented method, because it may be used incorrectly.*

        Set uncompressed content size of a frame, the size will be written into the
        frame header.
        1, If called when (.last_mode != .FLUSH_FRAME), a RuntimeError will be raised.
        2, If the actual size doesn't match the value, a ZstdError will be raised, and
           the last compressed chunk is likely to be lost.
        3, The size is only valid for one frame, then it restores to "unknown size".

        Parameter
        size: Uncompressed content size of a frame, None means "unknown size".
        """
        # Get size value
        if size is None:
            size = m.ZSTD_CONTENTSIZE_UNKNOWN
        else:
            try:
                if size < 0 or size > 2**64-1:
                    raise Exception
            except:
                msg = ("size argument should be 64-bit unsigned integer "
                       "value, or None.")
                raise ValueError(msg)

        with self._lock:
            # Check the current mode
            if self.__last_mode != m.ZSTD_e_end:
                msg = ("._set_pledged_input_size() method must be called "
                       "when (.last_mode == .FLUSH_FRAME).")
                raise RuntimeError(msg)

            # Set pledged content size
            zstd_ret = m.ZSTD_CCtx_setPledgedSrcSize(self._cctx, size)
            if m.ZSTD_isError(zstd_ret):
                _set_zstd_error(_ErrorType.ERR_SET_PLEDGED_INPUT_SIZE, zstd_ret)

    @property
    def last_mode(self):
        """The last mode used to this compressor object, its value can be .CONTINUE,
        .FLUSH_BLOCK, .FLUSH_FRAME. Initialized to .FLUSH_FRAME.

        It can be used to get the current state of a compressor, such as, data flushed,
        a frame ended.
        """
        return self.__last_mode

class RichMemZstdCompressor(_Compressor):
    """A compressor use rich memory mode. It is designed to allocate more memory,
    but faster in some cases.
    """

    def __init__(self, level_or_option=None, zstd_dict=None):
        """Initialize a RichMemZstdCompressor object.

        Parameters
        level_or_option: When it's an int object, it represents the compression level.
                         When it's a dict object, it contains advanced compression
                         parameters.
        zstd_dict:       A ZstdDict object, pre-trained zstd dictionary.
        """
        super().__init__(level_or_option=level_or_option, zstd_dict=zstd_dict)

        if self._use_multithread:
            msg = ('Currently "rich memory mode" has no effect on '
                   'zstd multi-threaded compression (set '
                   '"CParameter.nbWorkers" >= 1), it will allocate '
                   'unnecessary memory.')
            warn(msg, ResourceWarning, 1)

    def compress(self, data):
        """Compress data using rich memory mode, return a single zstd frame.

        Compressing b'' will get an empty content frame (9 bytes or more).

        Parameter
        data: A bytes-like object, data to be compressed.
        """
        with self._lock:
            try:
                ret = self._compress_impl(data, m.ZSTD_e_end, True)
                return ret
            except:
                # Resetting cctx's session never fail
                m.ZSTD_CCtx_reset(self._cctx, m.ZSTD_reset_session_only)
                raise



from threading import Lock

from .common import m, ffi, ZstdError, \
                    _set_zstd_error, _ErrorType

_DICT_TYPE_DIGESTED   = 0
_DICT_TYPE_UNDIGESTED = 1
_DICT_TYPE_PREFIX     = 2

class ZstdDict:
    """Zstd dictionary, used for compression/decompression."""

    def __init__(self, dict_content, is_raw=False):
        """Initialize a ZstdDict object.

        Parameters
        dict_content: A bytes-like object, dictionary's content.
        is_raw:       This parameter is for advanced user. True means dict_content
                      argument is a "raw content" dictionary, free of any format
                      restriction. False means dict_content argument is an ordinary
                      zstd dictionary, was created by zstd functions, follow a
                      specified format.
        """
        self.__cdicts = {}
        self.__ddict = ffi.NULL
        self.__lock = Lock()

        # Check dict_content's type
        try:
            self.__dict_content = bytes(dict_content)
        except:
            raise TypeError("dict_content argument should be bytes-like object.")

        # Both ordinary dictionary and "raw content" dictionary should
        # at least 8 bytes
        if len(self.__dict_content) < 8:
            raise ValueError('Zstd dictionary content should at least 8 bytes.')

        # Get dict_id, 0 means "raw content" dictionary.
        self.__dict_id = m.ZSTD_getDictID_fromDict(
                                    ffi.from_buffer(self.__dict_content),
                                    len(self.__dict_content))

        # Check validity for ordinary dictionary
        if not is_raw and self.__dict_id == 0:
            msg = ('The dict_content argument is not a valid zstd '
                   'dictionary. The first 4 bytes of a valid zstd dictionary '
                   'should be a magic number: b"\\x37\\xA4\\x30\\xEC".\n'
                   'If you are an advanced user, and can be sure that '
                   'dict_content argument is a "raw content" zstd '
                   'dictionary, set is_raw parameter to True.')
            raise ValueError(msg)

    def __del__(self):
        try:
            for level, cdict in self.__cdicts.items():
                m.ZSTD_freeCDict(cdict)
                self.__cdicts[level] = ffi.NULL
        except AttributeError:
            pass

        try:
            m.ZSTD_freeDDict(self.__ddict)
            self.__ddict = ffi.NULL
        except AttributeError:
            pass

    @property
    def dict_content(self):
        """The content of zstd dictionary, a bytes object, it's the same as dict_content
        argument in ZstdDict.__init__() method. It can be used with other programs.
        """
        return self.__dict_content

    @property
    def dict_id(self):
        """ID of zstd dictionary, a 32-bit unsigned int value.

        Non-zero means ordinary dictionary, was created by zstd functions, follow
        a specified format.

        0 means a "raw content" dictionary, free of any format restriction, used
        for advanced user.
        """
        return self.__dict_id

    @property
    def as_digested_dict(self):
        """Load as a digested dictionary to compressor, by passing this attribute as
        zstd_dict argument: compress(dat, zstd_dict=zd.as_digested_dict)
        1, Some advanced compression parameters of compressor may be overridden
           by parameters of digested dictionary.
        2, ZstdDict has a digested dictionaries cache for each compression level.
           It's faster when loading again a digested dictionary with the same
           compression level.
        3, No need to use this for decompression.
        """
        return (self, _DICT_TYPE_DIGESTED)

    @property
    def as_undigested_dict(self):
        """Load as an undigested dictionary to compressor, by passing this attribute as
        zstd_dict argument: compress(dat, zstd_dict=zd.as_undigested_dict)
        1, The advanced compression parameters of compressor will not be overridden.
        2, Loading an undigested dictionary is costly. If load an undigested dictionary
           multiple times, consider reusing a compressor object.
        3, No need to use this for decompression.
        """
        return (self, _DICT_TYPE_UNDIGESTED)

    @property
    def as_prefix(self):
        """Load as a prefix to compressor/decompressor, by passing this attribute as
        zstd_dict argument: compress(dat, zstd_dict=zd.as_prefix)
        1, Prefix is compatible with long distance matching, while dictionary is not.
        2, It only works for the first frame, then the compressor/decompressor will
           return to no prefix state.
        3, When decompressing, must use the same prefix as when compressing.
        """
        return (self, _DICT_TYPE_PREFIX)

    def __str__(self):
        return '<ZstdDict dict_id=%d dict_size=%d>' % \
               (self.__dict_id, len(self.__dict_content))

    def __len__(self):
        return len(self.__dict_content)

    def __reduce__(self):
        msg = ("ZstdDict object intentionally doesn't support pickle. If need "
               "to save zstd dictionary to disk, please save .dict_content "
               "attribute, it's a bytes object. So that the zstd dictionary "
               "can be used with other programs.")
        raise TypeError(msg)

    def _get_cdict(self, level):
        with self.__lock:
            # Already cached
            if level in self.__cdicts:
                cdict = self.__cdicts[level]
            else:
                # Create ZSTD_CDict instance
                cdict = m.ZSTD_createCDict(ffi.from_buffer(self.__dict_content),
                                           len(self.__dict_content), level)
                if cdict == ffi.NULL:
                    msg = ("Failed to create ZSTD_CDict instance from zstd "
                           "dictionary content. Maybe the content is corrupted.")
                    raise ZstdError(msg)
                self.__cdicts[level] = cdict
            return cdict

    def _get_ddict(self):
        # Already created
        if self.__ddict != ffi.NULL:
            return self.__ddict

        with self.__lock:
            # Create ZSTD_DDict instance from dictionary content
            self.__ddict = m.ZSTD_createDDict(
                                    ffi.from_buffer(self.__dict_content),
                                    len(self.__dict_content))

            if self.__ddict == ffi.NULL:
                msg = ("Failed to create ZSTD_DDict instance from zstd "
                       "dictionary content. Maybe the content is corrupted.")
                raise ZstdError(msg)

            return self.__ddict

def _load_c_dict(cctx, zstd_dict, level):
    if isinstance(zstd_dict, ZstdDict):
        # When compressing, use undigested dictionary by default.
        zd = zstd_dict
        type = _DICT_TYPE_UNDIGESTED
    elif isinstance(zstd_dict, tuple) and \
         len(zstd_dict) == 2 and \
         isinstance(zstd_dict[0], ZstdDict) and \
         zstd_dict[1] in {_DICT_TYPE_DIGESTED,
                          _DICT_TYPE_UNDIGESTED,
                          _DICT_TYPE_PREFIX}:
        zd = zstd_dict[0]
        type = zstd_dict[1]
    else:
        raise TypeError("zstd_dict argument should be ZstdDict object.")

    if type == _DICT_TYPE_DIGESTED:
        # Get ZSTD_CDict
        c_dict = zd._get_cdict(level)
        # Reference a prepared dictionary.
        # It overrides some compression context's parameters.
        zstd_ret = m.ZSTD_CCtx_refCDict(cctx, c_dict)
    elif type == _DICT_TYPE_UNDIGESTED:
        # Load a dictionary.
        # It doesn't override compression context's parameters.
        zstd_ret = m.ZSTD_CCtx_loadDictionary(
                                    cctx,
                                    ffi.from_buffer(zd.dict_content),
                                    len(zd.dict_content))
    elif type == _DICT_TYPE_PREFIX:
        # Reference as prefix
        zstd_ret = m.ZSTD_CCtx_refPrefix(
                                    cctx,
                                    ffi.from_buffer(zd.dict_content),
                                    len(zd.dict_content))
    else:
        raise SystemError('_load_c_dict() impossible code path')

    if m.ZSTD_isError(zstd_ret):
        _set_zstd_error(_ErrorType.ERR_LOAD_C_DICT, zstd_ret)

def _load_d_dict(dctx, zstd_dict):
    if isinstance(zstd_dict, ZstdDict):
        # When decompressing, use digested dictionary by default.
        zd = zstd_dict
        type = _DICT_TYPE_DIGESTED
    elif isinstance(zstd_dict, tuple) and \
         len(zstd_dict) == 2 and \
         isinstance(zstd_dict[0], ZstdDict) and \
         zstd_dict[1] in {_DICT_TYPE_DIGESTED,
                          _DICT_TYPE_UNDIGESTED,
                          _DICT_TYPE_PREFIX}:
        zd = zstd_dict[0]
        type = zstd_dict[1]
    else:
        raise TypeError("zstd_dict argument should be ZstdDict object.")

    if type == _DICT_TYPE_DIGESTED:
        # Get ZSTD_DDict
        d_dict = zd._get_ddict()
        # Reference a prepared dictionary
        zstd_ret = m.ZSTD_DCtx_refDDict(dctx, d_dict)
    elif type == _DICT_TYPE_UNDIGESTED:
        # Load a dictionary
        zstd_ret = m.ZSTD_DCtx_loadDictionary(
                                    dctx,
                                    ffi.from_buffer(zd.dict_content),
                                    len(zd.dict_content))
    elif type == _DICT_TYPE_PREFIX:
        # Reference as prefix
        zstd_ret = m.ZSTD_DCtx_refPrefix(
                                    dctx,
                                    ffi.from_buffer(zd.dict_content),
                                    len(zd.dict_content))
    else:
        raise SystemError('_load_d_dict() impossible code path')

    if m.ZSTD_isError(zstd_ret):
        _set_zstd_error(_ErrorType.ERR_LOAD_D_DICT, zstd_ret)



try:
    import cffi
except ImportError:
    # PyPy includes cffi by default
    msg = ('\n    To build the CFFI implementation '
           'of pyzstd module, need cffi module.'
           '\n    On CPython, CFFI implementation '
           'is slower than C implementation.\n')
    print(msg)

    # Used for PEP-517 get_requires_for_build_wheel hook
    from setuptools.build_meta import SetupRequirementsError
    raise SetupRequirementsError(['cffi'])

ffibuilder = cffi.FFI()

ffibuilder.cdef("""
#define ZSTD_VERSION_NUMBER ...
#define ZSTD_CONTENTSIZE_UNKNOWN ...
#define ZSTD_CONTENTSIZE_ERROR ...

typedef ... ZSTD_CDict;
typedef ... ZSTD_DDict;
typedef ... ZSTD_CCtx;
typedef ... ZSTD_DCtx;

typedef struct {
    size_t error;
    int lowerBound;
    int upperBound;
} ZSTD_bounds;

typedef enum {
    ZSTD_e_continue,
    ZSTD_e_flush,
    ZSTD_e_end
} ZSTD_EndDirective;

typedef struct ZSTD_inBuffer_s {
  const void* src;
  size_t size;
  size_t pos;
} ZSTD_inBuffer;

typedef struct ZSTD_outBuffer_s {
  void*  dst;
  size_t size;
  size_t pos;
} ZSTD_outBuffer;

typedef enum {
    /* Compression parameters */
    ZSTD_c_compressionLevel,

    /* Advanced compression parameters */
    ZSTD_c_windowLog,
    ZSTD_c_hashLog,
    ZSTD_c_chainLog,
    ZSTD_c_searchLog,
    ZSTD_c_minMatch,
    ZSTD_c_targetLength,
    ZSTD_c_strategy,

    /* LDM mode parameters */
    ZSTD_c_enableLongDistanceMatching,
    ZSTD_c_ldmHashLog,
    ZSTD_c_ldmMinMatch,
    ZSTD_c_ldmBucketSizeLog,
    ZSTD_c_ldmHashRateLog,

    /* frame parameters */
    ZSTD_c_contentSizeFlag,
    ZSTD_c_checksumFlag,
    ZSTD_c_dictIDFlag,

    /* multi-threading parameters */
    ZSTD_c_nbWorkers,
    ZSTD_c_jobSize,
    ZSTD_c_overlapLog,
} ZSTD_cParameter;

typedef enum {
    ZSTD_d_windowLogMax,
} ZSTD_dParameter;

typedef enum {
    ZSTD_fast,
    ZSTD_dfast,
    ZSTD_greedy,
    ZSTD_lazy,
    ZSTD_lazy2,
    ZSTD_btlazy2,
    ZSTD_btopt,
    ZSTD_btultra,
    ZSTD_btultra2
} ZSTD_strategy;

typedef enum {
    ZSTD_reset_session_only,
    ...
} ZSTD_ResetDirective;

size_t      ZSTD_compressBound(size_t srcSize);
unsigned    ZSTD_isError(size_t code);
const char* ZSTD_getErrorName(size_t code);
int         ZSTD_minCLevel(void);
int         ZSTD_maxCLevel(void);
int         ZSTD_defaultCLevel(void);

unsigned long long ZSTD_getFrameContentSize(const void *src, size_t srcSize);
unsigned ZSTD_getDictID_fromFrame(const void* src, size_t srcSize);
size_t ZSTD_findFrameCompressedSize(const void* src, size_t srcSize);

unsigned ZSTD_versionNumber(void);
const char* ZSTD_versionString(void);

ZSTD_bounds ZSTD_cParam_getBounds(ZSTD_cParameter cParam);
ZSTD_bounds ZSTD_dParam_getBounds(ZSTD_dParameter dParam);
size_t ZSTD_CCtx_setParameter(ZSTD_CCtx* cctx, ZSTD_cParameter param, int value);
size_t ZSTD_DCtx_setParameter(ZSTD_DCtx* dctx, ZSTD_dParameter param, int value);

size_t ZSTD_CStreamInSize(void);
size_t ZSTD_CStreamOutSize(void);
size_t ZSTD_DStreamInSize(void);
size_t ZSTD_DStreamOutSize(void);

ZSTD_CDict* ZSTD_createCDict(const void* dictBuffer, size_t dictSize,
                             int compressionLevel);
size_t ZSTD_CCtx_refCDict(ZSTD_CCtx* cctx, const ZSTD_CDict* cdict);
size_t ZSTD_CCtx_loadDictionary(ZSTD_CCtx* cctx, const void* dict, size_t dictSize);
size_t ZSTD_freeCDict(ZSTD_CDict* CDict);
size_t ZSTD_CCtx_refPrefix(ZSTD_CCtx* cctx, const void* prefix, size_t prefixSize);

ZSTD_DDict* ZSTD_createDDict(const void* dictBuffer, size_t dictSize);
size_t ZSTD_DCtx_refDDict(ZSTD_DCtx* dctx, const ZSTD_DDict* ddict);
size_t ZSTD_DCtx_loadDictionary(ZSTD_DCtx* dctx, const void* dict, size_t dictSize);
size_t ZSTD_freeDDict(ZSTD_DDict* ddict);
size_t ZSTD_DCtx_refPrefix(ZSTD_DCtx* dctx, const void* prefix, size_t prefixSize);

unsigned ZSTD_getDictID_fromDict(const void* dict, size_t dictSize);

ZSTD_CCtx* ZSTD_createCCtx(void);
size_t ZSTD_freeCCtx(ZSTD_CCtx* cctx);
size_t ZSTD_CCtx_reset(ZSTD_CCtx* cctx, ZSTD_ResetDirective reset);
size_t ZSTD_CCtx_setPledgedSrcSize(ZSTD_CCtx* cctx, unsigned long long pledgedSrcSize);
size_t ZSTD_compressStream2(ZSTD_CCtx* cctx,
                            ZSTD_outBuffer* output,
                            ZSTD_inBuffer* input,
                            ZSTD_EndDirective endOp);

ZSTD_DCtx* ZSTD_createDCtx(void);
size_t ZSTD_freeDCtx(ZSTD_DCtx* dctx);
size_t ZSTD_DCtx_reset(ZSTD_DCtx* dctx, ZSTD_ResetDirective reset);
size_t ZSTD_decompressStream(ZSTD_DCtx* dctx,
                             ZSTD_outBuffer* output,
                             ZSTD_inBuffer* input);

unsigned ZDICT_isError(size_t errorCode);
size_t ZDICT_trainFromBuffer(void* dictBuffer, size_t dictBufferCapacity,
                             const void* samplesBuffer,
                             const size_t* samplesSizes, unsigned nbSamples);


typedef struct {
    int      compressionLevel;
    unsigned notificationLevel;
    unsigned dictID;
} ZDICT_params_t;

size_t ZDICT_finalizeDictionary(void* dstDictBuffer, size_t maxDictSize,
                                const void* dictContent, size_t dictContentSize,
                                const void* samplesBuffer, const size_t* samplesSizes, unsigned nbSamples,
                                ZDICT_params_t parameters);

extern int pyzstd_static_link;
""")

source = """
#include "zstd.h"
#include "zdict.h"

#if ZSTD_VERSION_NUMBER < 10400
    #error "pyzstd module requires zstd v1.4.0+"
#endif

#if ZSTD_VERSION_NUMBER < 10405
typedef struct {
    int      compressionLevel;
    unsigned notificationLevel;
    unsigned dictID;
} ZDICT_params_t;

size_t ZDICT_finalizeDictionary(void* dstDictBuffer, size_t maxDictSize,
                                const void* dictContent, size_t dictContentSize,
                                const void* samplesBuffer, const size_t* samplesSizes, unsigned nbSamples,
                                ZDICT_params_t parameters)
{
    return 0;
}
#endif

#if ZSTD_VERSION_NUMBER < 10500
int ZSTD_defaultCLevel(void)
{
    return ZSTD_CLEVEL_DEFAULT;
}
#endif

#ifdef PYZSTD_STATIC_LINK
int pyzstd_static_link = 1;
#else
int pyzstd_static_link = 0;
#endif
"""

def get_extension(**kwargs):
    ffibuilder.set_source(source=source, **kwargs)
    return ffibuilder.distutils_extension()



from .common import m, ffi, ZstdError, _new_nonzero, \
                    _set_c_parameters, _set_d_parameters, \
                    _set_zstd_error, _ErrorType, \
                    _write_to_fp
from .dict import _load_c_dict, _load_d_dict

def _invoke_callback(callback, in_mv, in_buf, callback_read_pos,
                     out_mv, out_buf, total_input_size, total_output_size):
    # Only yield input data once
    in_size = in_buf.size - callback_read_pos
    callback_read_pos = in_buf.size

    # Don't yield empty data
    if in_size == 0 and out_buf.pos == 0:
        return callback_read_pos

    # memoryview
    in_memoryview = in_mv[:in_size]
    out_memoryview = out_mv[:out_buf.pos]

    # Callback
    callback(total_input_size, total_output_size,
             in_memoryview, out_memoryview)

    return callback_read_pos

def compress_stream(input_stream, output_stream, *,
                    level_or_option = None, zstd_dict = None,
                    pledged_input_size = None,
                    read_size = m.ZSTD_CStreamInSize(),
                    write_size = m.ZSTD_CStreamOutSize(),
                    callback = None):
    """Compresses input_stream and writes the compressed data to output_stream, it
    doesn't close the streams.

    If input stream is b'', nothing will be written to output stream.

    Return a tuple, (total_input, total_output), the items are int objects.

    Parameters
    input_stream: Input stream that has a .readinto(b) method.
    output_stream: Output stream that has a .write(b) method. If use callback
        function, this parameter can be None.
    level_or_option: When it's an int object, it represents the compression
        level. When it's a dict object, it contains advanced compression
        parameters.
    zstd_dict: A ZstdDict object, pre-trained zstd dictionary.
    pledged_input_size: If set this parameter to the size of input data, the
        size will be written into the frame header. If the actual input data
        doesn't match it, a ZstdError will be raised.
    read_size: Input buffer size, in bytes.
    write_size: Output buffer size, in bytes.
    callback: A callback function that accepts four parameters:
        (total_input, total_output, read_data, write_data), the first two are
        int objects, the last two are readonly memoryview objects.
    """
    level = 0  # 0 means use zstd's default compression level
    use_multithread = False
    total_input_size = 0
    total_output_size = 0

    # Check arguments
    if not hasattr(input_stream, "readinto"):
        raise TypeError("input_stream argument should have a .readinto(b) method.")
    if output_stream is not None:
        if not hasattr(output_stream, "write"):
            raise TypeError("output_stream argument should have a .write(b) method.")
    else:
        if callback is None:
            msg = ("At least one of output_stream argument and "
                   "callback argument should be non-None.")
            raise TypeError(msg)

    try:
        if read_size <= 0 or write_size <= 0:
            raise Exception
    except:
        msg = ("read_size argument and write_size argument should "
               "be positive numbers.")
        raise ValueError(msg)

    if pledged_input_size is not None:
        try:
            if pledged_input_size < 0 or pledged_input_size > 2**64-1:
                raise Exception
        except:
            msg = ("pledged_input_size argument should be 64-bit "
                   "unsigned integer value.")
            raise ValueError(msg)

    try:
        # Initialize & set ZstdCompressor
        cctx = m.ZSTD_createCCtx()
        if cctx == ffi.NULL:
            raise ZstdError("Unable to create ZSTD_CCtx instance.")

        if level_or_option is not None:
            level, use_multithread = \
                _set_c_parameters(cctx, level_or_option)
        if zstd_dict is not None:
            _load_c_dict(cctx, zstd_dict, level)

        if pledged_input_size is not None:
            zstd_ret = m.ZSTD_CCtx_setPledgedSrcSize(cctx, pledged_input_size)
            if m.ZSTD_isError(zstd_ret):
                _set_zstd_error(_ErrorType.ERR_COMPRESS, zstd_ret)

        # Input buffer, in.size and in.pos will be set later.
        in_buf = _new_nonzero("ZSTD_inBuffer *")
        if in_buf == ffi.NULL:
            raise MemoryError

        _input_block = ffi.buffer(_new_nonzero("char[]", read_size))
        in_mv = memoryview(_input_block)
        in_buf.src = ffi.from_buffer(_input_block)

        # Output buffer, out.pos will be set later.
        out_buf = _new_nonzero("ZSTD_outBuffer *")
        if out_buf == ffi.NULL:
            raise MemoryError

        _output_block = ffi.buffer(_new_nonzero("char[]", write_size))
        out_mv = memoryview(_output_block)
        out_buf.dst = ffi.from_buffer(_output_block)
        out_buf.size = write_size

        # Read
        while True:
            # Invoke .readinto() method
            read_bytes = input_stream.readinto(_input_block)
            if read_bytes < 0 or read_bytes > read_size:
                msg = ("input_stream.readinto() returned invalid length "
                       "%d (should be 0 <= value <= %d)") % \
                       (read_bytes, read_size)
                raise ValueError(msg)

            # Don't generate empty frame
            if read_bytes == 0 and total_input_size == 0:
                break
            total_input_size += read_bytes

            in_buf.size = read_bytes
            in_buf.pos = 0
            callback_read_pos = 0
            end_directive = m.ZSTD_e_end \
                            if (read_bytes == 0) \
                            else m.ZSTD_e_continue

            # Compress & write
            while True:
                # Output position
                out_buf.pos = 0

                # Compress
                if use_multithread and end_directive == m.ZSTD_e_continue:
                    while True:
                        zstd_ret = m.ZSTD_compressStream2(cctx, out_buf, in_buf, m.ZSTD_e_continue)
                        if (out_buf.pos == out_buf.size
                              or in_buf.pos == in_buf.size
                              or m.ZSTD_isError(zstd_ret)):
                            break
                else:
                    zstd_ret = m.ZSTD_compressStream2(cctx, out_buf, in_buf, end_directive)

                if m.ZSTD_isError(zstd_ret):
                    _set_zstd_error(_ErrorType.ERR_COMPRESS, zstd_ret)

                # Accumulate output bytes
                total_output_size += out_buf.pos

                # Write all output to output_stream
                if output_stream is not None:
                    _write_to_fp("output_stream.write()", output_stream,
                                 out_mv, out_buf)

                # Invoke callback
                if callback is not None:
                    callback_read_pos = _invoke_callback(
                                     callback, in_mv, in_buf, callback_read_pos,
                                     out_mv, out_buf, total_input_size, total_output_size)

                # Finished
                if use_multithread and end_directive == m.ZSTD_e_continue:
                    if in_buf.pos == in_buf.size and \
                       out_buf.pos != out_buf.size:
                        break
                else:
                    if zstd_ret == 0:
                        break

            # Input stream ended
            if read_bytes == 0:
                break

        return (total_input_size, total_output_size)
    finally:
        m.ZSTD_freeCCtx(cctx)

def decompress_stream(input_stream, output_stream, *,
                      zstd_dict = None, option = None,
                      read_size = m.ZSTD_DStreamInSize(),
                      write_size = m.ZSTD_DStreamOutSize(),
                      callback = None):
    """Decompresses input_stream and writes the decompressed data to output_stream,
    it doesn't close the streams.

    Supports multiple concatenated frames.

    Return a tuple, (total_input, total_output), the items are int objects.

    Parameters
    input_stream: Input stream that has a .readinto(b) method.
    output_stream: Output stream that has a .write(b) method. If use callback
        function, this parameter can be None.
    zstd_dict: A ZstdDict object, pre-trained zstd dictionary.
    option: A dict object, contains advanced decompression parameters.
    read_size: Input buffer size, in bytes.
    write_size: Output buffer size, in bytes.
    callback: A callback function that accepts four parameters:
        (total_input, total_output, read_data, write_data), the first two are
        int objects, the last two are readonly memoryview objects.
    """
    at_frame_edge = True
    total_input_size = 0
    total_output_size = 0

    # Check arguments
    if not hasattr(input_stream, "readinto"):
        raise TypeError("input_stream argument should have a .readinto(b) method.")
    if output_stream is not None:
        if not hasattr(output_stream, "write"):
            raise TypeError("output_stream argument should have a .write(b) method.")
    else:
        if callback is None:
            msg = ("At least one of output_stream argument and "
                   "callback argument should be non-None.")
            raise TypeError(msg)

    try:
        if read_size <= 0 or write_size <= 0:
            raise Exception
    except:
        msg = ("read_size argument and write_size argument should "
               "be positive numbers.")
        raise ValueError(msg)

    try:
        # Initialize & set ZstdDecompressor
        dctx = m.ZSTD_createDCtx()
        if dctx == ffi.NULL:
            raise ZstdError("Unable to create ZSTD_DCtx instance.")

        if zstd_dict is not None:
            _load_d_dict(dctx, zstd_dict)
        if option is not None:
            _set_d_parameters(dctx, option)

        # Input buffer, in.size and in.pos will be set later.
        in_buf = _new_nonzero("ZSTD_inBuffer *")
        if in_buf == ffi.NULL:
            raise MemoryError

        _input_block = ffi.buffer(_new_nonzero("char[]", read_size))
        in_mv = memoryview(_input_block)
        in_buf.src = ffi.from_buffer(_input_block)

        # Output buffer, out.pos will be set later.
        out_buf = _new_nonzero("ZSTD_outBuffer *")
        if out_buf == ffi.NULL:
            raise MemoryError

        _output_block = ffi.buffer(_new_nonzero("char[]", write_size))
        out_mv = memoryview(_output_block)
        out_buf.dst = ffi.from_buffer(_output_block)
        out_buf.size = write_size

        # Read
        while True:
            # Invoke .readinto() method
            read_bytes = input_stream.readinto(_input_block)
            if read_bytes < 0 or read_bytes > read_size:
                msg = ("input_stream.readinto() returned invalid length "
                       "%d (should be 0 <= value <= %d)") % \
                       (read_bytes, read_size)
                raise ValueError(msg)

            total_input_size += read_bytes

            in_buf.size = read_bytes
            in_buf.pos = 0
            callback_read_pos = 0

            # Decompress & write
            while True:
                # AFE check for setting .at_frame_edge flag, search "AFE" in
                # decompressor.c to see details.
                if at_frame_edge and in_buf.pos == in_buf.size:
                    break

                # Output position
                out_buf.pos = 0

                # Decompress
                zstd_ret = m.ZSTD_decompressStream(dctx, out_buf, in_buf)
                if m.ZSTD_isError(zstd_ret):
                    _set_zstd_error(_ErrorType.ERR_DECOMPRESS, zstd_ret)

                # Set .af_frame_edge flag
                at_frame_edge = (zstd_ret == 0)

                # Accumulate output bytes
                total_output_size += out_buf.pos

                # Write all output to output_stream
                if output_stream is not None:
                    _write_to_fp("output_stream.write()", output_stream,
                                 out_mv, out_buf)

                # Invoke callback
                if callback is not None:
                    callback_read_pos = _invoke_callback(
                                     callback, in_mv, in_buf, callback_read_pos,
                                     out_mv, out_buf, total_input_size, total_output_size)

                # Finished. When a frame is fully decoded, but not fully flushed,
                # the last byte is kept as hostage, it will be released when all
                # output is flushed.
                if in_buf.pos == in_buf.size:
                    # If input stream ends in an incomplete frame, output as
                    # much as possible.
                    if (read_bytes == 0
                          and not at_frame_edge
                          and out_buf.pos == out_buf.size):
                        continue

                    break

            # Input stream ended
            if read_bytes == 0:
                # Check data integrity. at_frame_edge flag is 1 when both the
                # input and output streams are at a frame edge.
                if not at_frame_edge:
                    msg = ("Decompression failed: zstd data ends in an "
                           "incomplete frame, maybe the input data was "
                           "truncated. Total input %d bytes, total output "
                           "%d bytes.") % (total_input_size, total_output_size)
                    raise ZstdError(msg)
                break

        return (total_input_size, total_output_size)
    finally:
        m.ZSTD_freeDCtx(dctx)



#!/usr/bin/env python3
import fnmatch
import os
import platform
import re
import sys
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

def read_stuff():
    ROOT_PATH = os.path.dirname(os.path.realpath(__file__))

    # read README.rst
    README_PATH = os.path.join(ROOT_PATH, 'README.rst')
    with open(README_PATH, 'r', encoding='utf-8') as file:
        long_description = file.read()

    # read module version
    INIT_PATH = os.path.join(ROOT_PATH, 'src', '__init__.py')
    with open(INIT_PATH, 'r', encoding='utf-8') as file:
        file_content = file.read()
        m = re.search(r'''__version__\s*=\s*(['"])(.*?)\1''', file_content)
        module_version = m.group(2)

    return long_description, module_version

def get_zstd_files_list():
    ret = []
    for sub_dir in ('common', 'compress', 'decompress', 'dictBuilder'):
        directory = 'zstd/lib/' + sub_dir + '/'
        dir_list = os.listdir(directory)

        # Source files
        l = [directory + fn
               for fn in dir_list
               if fnmatch.fnmatch(fn, '*.[cCsS]')]
        ret.extend(l)
    return ret

def has_option(option):
    if option in sys.argv:
        print(' * build pyzstd wheel with option:', option)
        sys.argv = [s for s in sys.argv if s != option]
        return True
    else:
        return False

class pyzstd_build_ext(build_ext):
    PYZSTD_AVX2 = False
    PYZSTD_DEBUG = False
    PYZSTD_WARNING_AS_ERROR = False
    PYZSTD_CONFIG_MSG = ''

    def build_extensions(self):
        # Print build config message in actual build
        print(self.PYZSTD_CONFIG_MSG)

        # Accept assembly files
        self.compiler.src_extensions.extend(['.s', '.S'])
        # Build debug build
        self.debug = self.PYZSTD_DEBUG

        for extension in self.extensions:
            if self.compiler.compiler_type in ('unix', 'mingw32', 'cygwin'):
                # -g0:
                #   Level 0 produces no debug information at all. This reduces
                #   the size of GCC wheels. By default CPython won't print any
                #   C stack trace, so -g0 and -g2 are same for most users.
                # -flto:
                #   This option runs the standard link-time optimizer. To use the
                #   link-time optimizer, -flto and optimization options should be
                #   specified at compile time and during the final link.
                more_options = ['-g0', '-flto']
                if self.PYZSTD_AVX2:
                    instrs = ['-mavx2', '-mlzcnt', '-mbmi', '-mbmi2']
                    more_options.extend(instrs)
                if self.PYZSTD_WARNING_AS_ERROR:
                    more_options.append('-Werror')
                extension.extra_compile_args.extend(more_options)
                extension.extra_link_args.extend(['-g0', '-flto'])
            elif self.compiler.compiler_type == 'msvc':
                # Remove .S source files, they use gcc/clang syntax.
                extension.sources = [i for i in extension.sources
                                        if not fnmatch.fnmatch(i, '*.[sS]')]

                # /Ob3: More aggressive inlining than /Ob2.
                # /GF:  Eliminates duplicate strings.
                # /Gy:  Does function level linking.
                #   /Ob3 is a bit faster on the whole. In setuptools v56.1+,
                #   /GF and /Gy are enabled by default, they reduce the size
                #   of MSVC wheels.
                more_options = ['/Ob3', '/GF', '/Gy']
                if self.PYZSTD_AVX2:
                    more_options.append('/arch:AVX2')
                if self.PYZSTD_WARNING_AS_ERROR:
                    more_options.append('/WX')
                extension.extra_compile_args.extend(more_options)
        super().build_extensions()

def do_setup():
    # read stuff
    long_description, module_version = read_stuff()

    # parse options
    pyzstd_build_ext.PYZSTD_AVX2 = has_option('--avx2')
    pyzstd_build_ext.PYZSTD_DEBUG = has_option('--debug')
    pyzstd_build_ext.PYZSTD_WARNING_AS_ERROR = has_option('--warning-as-error')

    DYNAMIC_LINK = has_option('--dynamic-link-zstd')
    CFFI = has_option('--cffi') or platform.python_implementation() == 'PyPy'
    MULTI_PHASE_INIT = has_option('--multi-phase-init')

    # build config message
    pyzstd_build_ext.PYZSTD_CONFIG_MSG = \
               ('+--------------------------------------------+\n'
                '|             Pyzstd build config            |\n'
                '+-------------------------+------------------+\n'
                '| Pyzstd version          | {!s:<16} |\n'
                '+-------------------------+------------------+\n'
                '| Implementation          | {!s:<16} |\n'
                '+-------------------------+------------------+\n'
                '| Enable multi-phase-init | {!s:<16} |\n'
                '+-------------------------+------------------+\n'
                '| Link to zstd library    | {!s:<16} |\n'
                '+-------------------------+------------------+\n'
                '| Enable AVX2/BMI2        | {!s:<16} |\n'
                '+-------------------------+------------------+\n'
                '| Debug build             | {!s:<16} |\n'
                '+-------------------------+------------------+\n'
                '| Warning as error        | {!s:<16} |\n'
                '+-------------------------+------------------+').format(
                    module_version,
                    'CFFI' if CFFI else 'C',
                    'Not for CFFI' if CFFI else MULTI_PHASE_INIT,
                    'Dynamically link' if DYNAMIC_LINK else 'Statically link',
                    pyzstd_build_ext.PYZSTD_AVX2,
                    pyzstd_build_ext.PYZSTD_DEBUG,
                    pyzstd_build_ext.PYZSTD_WARNING_AS_ERROR)

    if DYNAMIC_LINK:
        kwargs = {
            'include_dirs': [],    # .h directory
            'library_dirs': [],    # .lib directory
            'libraries': ['zstd'], # lib name, not filename, for the linker.
            'sources': [],
            'define_macros': []
        }
    else:  # statically link to zstd lib
        kwargs = {
            'include_dirs': ['zstd/lib',
                             # for zstd 1.4.x:
                             'zstd/lib/common',
                             'zstd/lib/dictBuilder'],
            'library_dirs': [],
            'libraries': [],
            'sources': get_zstd_files_list(),
            'define_macros': [('PYZSTD_STATIC_LINK', None),
                              # enable multi-threaded compression
                              ('ZSTD_MULTITHREAD', None)]
        }

    if CFFI:
        # packages
        packages = ['pyzstd', 'pyzstd.cffi']

        # binary extension
        kwargs['module_name'] = 'pyzstd.cffi._cffi_zstd'

        sys.path.append('src/bin_ext')
        import pyzstd_build_cffi
        binary_extension = pyzstd_build_cffi.get_extension(**kwargs)
    else:  # C implementation
        # packages
        packages = ['pyzstd', 'pyzstd.c']

        # binary extension
        kwargs['name'] = 'pyzstd.c._zstd'
        kwargs['sources'].append('src/bin_ext/pyzstd.c')
        if MULTI_PHASE_INIT:
            # use multi-phase initialization (PEP-489) on CPython 3.11+
            kwargs['define_macros'].append(('USE_MULTI_PHASE_INIT', None))

        binary_extension = Extension(**kwargs)

    setup(
        name='pyzstd',
        version=module_version,
        description=("Python bindings to Zstandard (zstd) compression "
                     "library, the API style is similar to Python's "
                     "bz2/lzma/zlib modules."),
        long_description=long_description,
        long_description_content_type='text/x-rst',
        author='Ma Lin',
        author_email='malincns@163.com',
        url='https://github.com/animalize/pyzstd',
        license='The 3-Clause BSD License',
        python_requires='>=3.5',

        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Topic :: System :: Archiving :: Compression",
            "License :: OSI Approved :: BSD License",
            "Programming Language :: Python :: Implementation :: CPython",
            "Programming Language :: Python :: Implementation :: PyPy",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
        ],
        keywords='zstandard zstd zst tar file seekable format',

        package_dir={'pyzstd': 'src'},
        packages=packages,
        package_data={'pyzstd': ['__init__.pyi', 'py.typed']},

        ext_modules=[binary_extension],
        cmdclass={'build_ext': pyzstd_build_ext},

        test_suite='tests'
    )

if __name__ == '__main__':
    do_setup()
